/****************************************************************************/
/* CRIBCORE32.C contains core decision-making routines for game of cribbage.*/
/****************************************************************************/
// 02 Jan 2019 - Craig Hessel.  This is 32-bit update of 16-bit CRIBCORE.C for 
// VB3 program Cribbage '97.  Update has identical functionality to 1997
// routines and is called from VB6 CRIBBAGE32.EXE.
//
// 1. LibMain replaced by DllMain
// 2. Export declarations modified to use _stdcall
// 3. Included ole2.h
// 4. Globally replaced int declarations with short declarations
// 5. Modified any long declarations found
// 6. Painfully debugged - ToPower required two added WORD casts
// 7. Some minor changes - constant arrays now uniformly BYTE or WORD
// 8. Omitted an export no longer needed in VB6 calling app
// 
// 21 Jun 2021 - Craig Hessel.  This update improves the decision-making of
// the 2019 routines.  Although labelled above as CRIBCORE32.C, this file
// is actually CRIBALT32.C and compiles as CRIBALT32.DLL in order to simplify
// testing.  File and DEF file have all 6 exports required of CRIBCORE32.DLL
// although only 2 specific exports are needed for the alternate DLL.  
//
// 1. Removed unneeded ole2.h header and FAR declarations, and converted
//      (with testing) most short declarations back to int
// 2. Included C file with two constant packed tables from GEN.C
// 3. Added 4 global static tables, initialized in DllMain
// 4. Adapted routines from DS.C (based on GEN.C) to replace static discard 
//      average tables with varying tables that account for 6 known cards 
// 5. Replaced static 4-card hand frequency tables with varying versions
//      that account for 7 known cards (also using GEN.C data)
// 6. Replaced bit/byte manipulation in risk assessment with double precision
//      handling, including log(), pow(), and cos() calls from math library
// 7. After further revamping, risk assessment method still needs work, as 
//      program barely relies on it now
//  
// Tuning of many program equates (mostly related to pegging/risk) was done 
// by repeatedly using seed 13896 in 1000-game run pairs with first deals 
// alternating.  2019 routines competed as opponent against 2021 routines as 
// computer in one run.  In next run, roles were reversed.  Values below show 
// a 36-game net positive swing over the two runs. Results are game, 2-3-4 
// match point, and total point differentials. Opp means 2021 code making 
// opponent decisions.  Com means 2021 code making computer decisions.
//
// Tuning results:  Opp +30/+74/+894  Com +42/+77/+474  Net +72/+151/+1368
//
// Following are post-tuning comparison test results using various other 
// random seeds in pairs of 1000-game runs.  Positive net is win for 2021 
// code.  Negative net is win for 2019 (i.e., 1997) code.  See raw results
// in file PostTuneFiveSeedTests.log.  
//
// Seed 314159265  Opp +30/ +66/ +904  Com +16/+33/ +710  Net +46/ +99/+1614
// Seed 271828182  Opp  +2/  +1/ +198  Com +14/+56/+1073  Net +16/ +57/+1271
// Seed 161803398  Opp +66/+159/+1547  Com +24/+42/ +180  Net +90/+201/+1727
// Seed 141421356  Opp  +2/   0/ +528  Com  +2/+15/ +876  Net  +4/ +15/+1404
// Seed 173205080  Opp  -6/ -27/  -62  Com +30/+82/+1148  Net +24/ +55/+1086
//
// New code clearly is scoring more (0.55 to 0.85 points per game above), due
// to use of DS.C discarding method.  Pegging difference is not clear.
 
#include <windows.h>
#include <math.h>     /* included for log(), pow(), cos() */
#include <stdio.h>    /* needed only for debugging file writes */

#define SuitOf(Card)  ((Card) & 0xF0)
#define RankOf(Card)  ((Card) & 0x0F)
#define CountOf(Rank) ((Rank) > 10 ? 10 : (Rank))
 
/* used in many pegging functions */
typedef struct tagPegStruct
  {int    *CtPtr;   /* points one before 13-entry card count array */
   double *WtPtr;   /* points one before 13-entry card weight array */
   LPSTR   PlPtr;   /* points into 8-byte played card array */
   LPSTR   HnPtr;   /* points to 11-byte own/cut/opp card array */
   int     GoFlgs;  /* 2 bits--set bit flags go pending */
   int     OppMin;  /* known opponent minimum rank from prior go */
   int     HnFlgs;  /* 4 bits--set bit flags own card available */
   int     OppLft;  /* number of opponent cards left */
   int     PegCnt;  /* current peg count */
   int     PlCnt;   /* count of played cards from last restart */
   int     Depth;   /* recursion depth (opponent plays to weigh) */
   double  Risk;    /* risk factor [-1.0,1.0] for push/balk */
   int     Bits;    /* or'ed 8/4/2/1 flags (peg plays) or 0 (go) */
  } PEGSTRUCT;

typedef PEGSTRUCT * LPPEGSTRUCT;

/* used only in ChkPeg() and BuildPl() */
typedef struct tagHandStruct
  {LPSTR  Hand;
   int    Len;
   int    Ndx;
  } HANDSTRUCT;

typedef HANDSTRUCT * LPHANDSTRUCT;

/* include 2 fixed tables and define 5 constants from GEN.C output: */
/* const unsigned char PnPack[43181, DlPack[44349] */
#include "2Tbls.c"

#define PN_SAMESUITREDUCTION (0.8954)
#define DL_SAMESUITREDUCTION (0.9603)
#define PN_HANDAVERAGE       (8.1297)
#define DL_HANDAVERAGE       (7.9804)
#define CB_HANDAVERAGE       (4.7816)

#define SUM45   (8145060) /* (45 6) */
#define RNKSIZE (18395)   

#define ASIZE   (1716*15)                                 /*  25740 */ 
#define BSIZE   (6435*11)                                 /*  70785 */
#define CSIZE   (4290*8 + 2860*7)                         /*  54340 */
#define DSIZE   (1716*5 + 858*4 + 286*6)                  /*  13728 */
#define ESIZE   (156*3 + 78*3)                            /*    702 */
#define TOTSIZE (ASIZE + BSIZE + CSIZE + DSIZE + ESIZE)   /* 165295 */

/* array referenced in LoadTot() and LoadHnTot() */
const int Wgts[] = {0,0,0,0,0,     /* 0 cards available */
                    1,1,0,0,0,
                    1,2,1,0,0,
                    1,3,3,1,0,
                    1,4,6,4,1};    /* 4 cards available */

/* arrays referenced only in SetNdx() */
const int QkWts2[] = {0,1,3,6, 10,15, 21, 28, 36, 45, 55,  66,  78};
const int QkWts1[] = {0,1,4,10,20,35, 56, 84,120,165,220, 286, 364};
const int QkWts0[] = {0,1,5,15,35,70,126,210,330,495,715,1001,1365};

char  DsNdxTbl[TOTSIZE]; /* fixed discard indexes 0-90 for all hands by rank */
short HnNdxTbl[TOTSIZE]; /* fixed held hand indexes 0-1819 for all hands by rank */
short PnTbl[TOTSIZE];    /* fixed pone discard weights unpacked from PnPack[] */
short DlTbl[TOTSIZE];    /* fixed dealer discard weights unpacked from DlPack[] */

/* DLLMain declaration--replaces LibMain from 1997 */
BOOL WINAPI DllMain (HINSTANCE const instance, 
                     DWORD const reason, 
                     LPVOID const reserved);

/* six export (_stdcall) declarations plus two helper functions */
int    ArgErrMsg(LPSTR FuncName);
int    ReadHand(LPSTR Tgt, LPSTR Src, int NumCards);
short  _stdcall Discard(LPSTR Hand, WORD DlScr, WORD PnScr);
short  _stdcall ScoreHand(LPSTR Hand, short IsCrib);
short  _stdcall ParseHand(LPSTR Hand, short IsCrib);
short  _stdcall MakePegPlay(LPSTR Cards, WORD Length, WORD OwnLeft, short IsDealer, 
                            short IsRestart, WORD OwnScr, WORD OppScr);
short  _stdcall ScorePegPlay(LPSTR Cards, WORD Length);
int    _stdcall RandSeed(DWORD Seed);

/* initialization section declarations--see DLLMain() */
int    GetFac(int Ndx);
int    GetBits(int BitLen, const unsigned char Tbl[], int *BitNdx);
int    Unpack(short Tbl[], const unsigned char PkTbl[]);
int    GetMask(char Hand[]);
void   SetNdx(char Hand[], int *Ndx, int BitMask);
void   LoadNdx();

/* risk assessment section declarations */
int    GetThresh(int Score, int PegPoint);
double BoostRisk(double Risk, int LowThresh);
double EvalShape(int ThreshHold);
double EvalRisk(int DlScr, int PnScr);
double EndPlayRisk(int OwnPlayed, int OppPlayed, int OwnScr, int OppScr, 
                   int IsDealer, int HndScr);
double EvalPegRisk(char Cards[], int OwnPlayed, int OppPlayed, int OwnScr, 
                   int OppScr, int IsDealer);
double GetDisScale(double Risk);
double GetPegScale(double Risk);

/* discard section declarations */
int    Score5(char Hand[], int IsCrib, int IsParse);
int    Spread(char Hand[]);
void   BumpTot(int Wt, int PrevNdx, int MaxWt, int Len, int *Ndx, 
               double PnFreq[], double DlFreq[], short WtPrev[]);
void   LoadTot(char Hand[], int IncrementalLoad, 
               double PnFreq[], double DlFreq[],short WtPrev[]);
void   SetAve2(char Hand[], double PnAve[], double DlAve[]);
void   SetAve4(char Hand[], double HnAve[], int HnMn[], int HnMx[], 
               int Sprd[], int CbMin[]);
int    Round(double Val);
int    PickDis(char Hand[], int DlScr, int PnScr);

/* pegging section declarations */
int    ScorePeg(char Cards[], int Length);
int    ChkPeg(char Cards[], int Length, int CardsLeft, int IsOwnTurn, int IsRestart);
int    ChoosePeg(char Cards[], int OwnPlayed, int OppPlayed, int OwnScr, 
                 int OppScr, int IsDealer, int IsRestart);
int    ConvPeg(char Cards[], int Bits);
void   BuildCt(char Cards[], int Length, int CtAry[]);
void   BuildPl(LPSTR Cards, int OwnPlayed, int OppPlayed, int IsOwnTurn, LPSTR PlAry);
void   Replay(LPPEGSTRUCT PegPtr, int OwnPlayed, int OppPlayed, 
              int IsDealer, int IsRestart);
int    QuickChk(LPPEGSTRUCT PegPtr, int OwnPlayed, int PointsToWin);
void   BumpHnTot(int Wt, int MaxWt, int Len, int *Ndx, double HnFreq[], int IsDlr);
void   LoadHnTot(char Hand[], double HnFreq[], int IsDlr);
void   InitWts(double WtAry[], double FreqTbl[], char KnownCds[], 
               int Len, int RankMin);
void   AdjFlWts(double WtAry[], char Cards[]);
void   DropWts(double WtAry[], LPSTR PlAry, int Length, int PegCnt);
void   AdjustWts(double WtAry[], char OwnCards[], char Played[], int Length);
void   NormWts(double WtAry[], int RankMin);
double GetProb(double x, double y, int z);
double ProdDiv(double x, double y, double z);
double XsAdd(double x, double y);
double PickPeg(LPPEGSTRUCT PegPtr, int IsTopLevel);
double PegAve(PEGSTRUCT PegData);

/* debugging delarations--functions may be commented out later */
void   DebugDump(char Hand[], double HnFreq[], int IsDlr);
int    DebugWanted(LPPEGSTRUCT PegPtr);
void   DebugPlay(LPPEGSTRUCT PegPtr, double Ave[]);

/****************************************************************************/
/******************** DllMain/Helper/Export functions ***********************/
/****************************************************************************/

/*----------------------------------------------------------------------------
; DLL start. Called once for reason DLL_PROCESS_ATTACH.  Ignore other calls.
;---------------------------------------------------------------------------*/
BOOL WINAPI DllMain (
    HINSTANCE const instance,  /* handle to DLL module */
    DWORD     const reason,    /* reason for calling function */
    LPVOID    const reserved)  /* reserved */
  {if (reason == DLL_PROCESS_ATTACH)
     {
      /* unpack main arrays (bit count check-values come from GEN.C) */
      if (Unpack(PnTbl, PnPack) != 345442 || Unpack(DlTbl, DlPack) != 354789)
        {
         MessageBox(GetActiveWindow(), "Unpack error--possibly corrupt DLL",
                   "DLL Load Terminated", MB_ICONSTOP);
         return FALSE;
        }

      /* load static DsNdxTbl/HnNdxTbl tables with precomputed indexes */
      LoadNdx();
     }	
	return TRUE;
  }

/*----------------------------------------------------------------------------
; Display error message if bad arguments passed to export function by VB.
;---------------------------------------------------------------------------*/
int ArgErrMsg(LPSTR FuncName)
  {char Buf[48];

   sprintf(Buf, "CribCore32.dll: %s() argument error", FuncName);
   MessageBox(GetActiveWindow(), Buf, "Call Terminated", MB_ICONSTOP);
   return -1;
  }

/*----------------------------------------------------------------------------
; Convert VB hand to DLL card format and return total.count (255 if bad card).
;---------------------------------------------------------------------------*/
int ReadHand(LPSTR Tgt, LPSTR Src, int NumCards)
  {int n, m, Rank, Suit, Card, Cnt = 0;

   for (n = 0; n < NumCards; n++)
     {Card = Src[n];
      if (Card < 0 || Card > 51)                  return 255;
      for (m = 0; m < n; m++) if (Card == Src[m]) return 255;

      Rank = Card % 13 + 1;
      Suit = 16 << (Card / 13);
      Tgt[n] = (char)(Rank | Suit);
      Cnt += CountOf(Rank);
     }

   return Cnt;
  }

/*----------------------------------------------------------------------------
; Export: Return dealer and pone discard choices for VB.
;---------------------------------------------------------------------------*/
short _stdcall Discard(LPSTR Hand, WORD DlScr, WORD PnScr)
  {char Hn[6];

   /* construct DLL-format hand from passed hand */
   if (DlScr > 120 || PnScr > 120 || ReadHand(Hn, Hand, 6) == 255)
      return ArgErrMsg("Discard");

   /* return dealer/pone discard indices in high/low bytes */
   return PickDis(Hn, DlScr, PnScr);
  }

/*----------------------------------------------------------------------------
; Export: Score 5-card hand for VB.
;---------------------------------------------------------------------------*/
short _stdcall ScoreHand(LPSTR Hand, short IsCrib)
  {char Hn[5];

   /* construct DLL-format hand from passed hand */
   if (ReadHand(Hn, Hand, 5) == 255)
      return ArgErrMsg("ScoreHand");

   return Score5(Hn, IsCrib, 0);
  }

/*----------------------------------------------------------------------------
; Export: Parse 5-card hand score for descriptive text in VB.
;---------------------------------------------------------------------------*/
short _stdcall ParseHand(LPSTR Hand, short IsCrib)
  {char Hn[5];

   /* construct DLL-format hand from passed hand */
   if (ReadHand(Hn, Hand, 5) == 255)
      return ArgErrMsg("ParseHand");

   return Score5(Hn, IsCrib, 1);
  }

/*----------------------------------------------------------------------------
; Export: Return pegging play choice (or 0 for go) for VB.
;---------------------------------------------------------------------------*/
short _stdcall MakePegPlay(LPSTR Cards, WORD Length, WORD OwnLeft,
      short IsDealer, short IsRestart, WORD OwnScr, WORD OppScr)
  {char Hn[11];

   /* construct DLL-format cards from passed cards and verify legal-- */
   /* expect own played cards in order, then own cards left, then own */
   /* discard, then cut, then opponent played cards in order */
   if (OwnScr > 120 || OppScr > 120 || Length < 7 || Length > 11 ||
       OwnLeft > 4 || ReadHand(Hn, Cards, Length) == 255 ||
       ChkPeg(Hn, Length, OwnLeft, !IsDealer, IsRestart) )
      return ArgErrMsg("MakePegPlay");

   /* return index 1-4 as best play or 0 if go */
   return ChoosePeg(Hn, 4-OwnLeft, Length-7, OwnScr, OppScr, IsDealer, IsRestart);
  }

/*----------------------------------------------------------------------------
; Export: Score pegging play of last card played for VB.
;---------------------------------------------------------------------------*/
short _stdcall ScorePegPlay(LPSTR Cards, WORD Length)
  {char Hn[8];

   /* construct DLL-format cards from passed cards--sequence since reset */
   if (Length > 8 || ReadHand(Hn, Cards, Length) > 31)
      return ArgErrMsg("ScorePegPlay");

   return ScorePeg(Hn, Length);
  }

/*----------------------------------------------------------------------------
; Export: Return updated pseudo-random seed for VB card generation.
;---------------------------------------------------------------------------*/
int _stdcall RandSeed(DWORD Seed)
  {
   /* constant must end in ...101 binary */
   return (Seed + 1) * 0x711CB3A5;
  }

/****************************************************************************/
/********************* DllMain Initialization Routines **********************/
/****************************************************************************/

/*----------------------------------------------------------------------------
; Get multiply/divide factor for expansion/reduction.
;---------------------------------------------------------------------------*/
int GetFac(int Ndx)
  {int Factor, Mod;

   /* some region-A divisions have remainder 4 and/or result over 256 */
   if      (Ndx < ASIZE)             Factor = 12;
   else if (Ndx < ASIZE+BSIZE)       Factor = 12;
   else if (Ndx < ASIZE+BSIZE+CSIZE)
     {
      Mod = (Ndx-ASIZE-BSIZE) % (6*8 + 4*7);
      if      (Mod < 6*8)            Factor = 12;
      else                           Factor = 4;
     }
   else if (Ndx < ASIZE+BSIZE+CSIZE+DSIZE)
     {
      Mod = (Ndx-ASIZE-BSIZE-CSIZE) % (6*5 + 3*4 + 1*6);
      if      (Mod < 6*5)            Factor = 12;
      else if (Mod < 6*5 + 3*4)      Factor = 4;
      else                           Factor = 6;
     }
   else
     {
      Mod = (Ndx-ASIZE-BSIZE-CSIZE-DSIZE) % (2*3 + 1*3);
      if      (Mod < 2*3)            Factor = 6;
      else                           Factor = 4;
     }

   return Factor;
  }

/*----------------------------------------------------------------------------
; Get up to 31 bits from character array and update BitNdx.
;---------------------------------------------------------------------------*/
int GetBits(int BitLen, const unsigned char Tbl[], int *BitNdx)
  {int Ndx, NextBit, Bits;

   /* initialize loop parameters and update output index */
   Bits     = 0;
   Ndx      = *BitNdx >> 3;
   NextBit  = 0x80 >> (*BitNdx & 7);
   *BitNdx += BitLen;

   /* collect output bits */
   while (BitLen--)
     {
      if (Bits <<= 1, NextBit & Tbl[Ndx]) ++Bits;
      if ((NextBit >>= 1) == 0) (++Ndx, NextBit = 0x80);
     }

   return Bits;
  }

/*----------------------------------------------------------------------------
; Unpack TOTSIZE short integers from character array.  Hofmann 1-1-1-15-31
; compression was used, with some extra data-specific handling.
;---------------------------------------------------------------------------*/
int Unpack(short Tbl[], const unsigned char PkTbl[])
  {int n, m, t, BitNdx;

   /* clear target array */
   for (n = 0; n < TOTSIZE; n++) Tbl[n] = 0;

   /* point index past 1+1+1+15+31 header ids in table */
   BitNdx = 8*49;

   /* unpack data after header */
   for (n = 0; n < TOTSIZE; n++)
     {
      /* handle region-A 4/3072 flags before each block of 15 bytes */
      if (n < ASIZE && n % 15 == 0)
        {
         if (15 == (m = t = GetBits(4, PkTbl, &BitNdx)))
           (m = GetBits(4, PkTbl, &BitNdx), t = GetBits(4, PkTbl, &BitNdx));

         /* add flag values */
         Tbl[n+m] += 4;
         if (t < 15) Tbl[n+t] += 3072;
        }

      if      (    GetBits(1, PkTbl, &BitNdx)) t = PkTbl[0];
      else if (    GetBits(1, PkTbl, &BitNdx)) t = PkTbl[1];
      else if (    GetBits(1, PkTbl, &BitNdx)) t = PkTbl[2];
      else if (m = GetBits(4, PkTbl, &BitNdx)) t = PkTbl[m+2];
      else if (m = GetBits(5, PkTbl, &BitNdx)) t = PkTbl[m+17];
      else                                     t = GetBits(8, PkTbl, &BitNdx);

      /* add expanded byte */
      Tbl[n] += t * GetFac(n);
     }

   /* return total number of bits scanned */
   return BitNdx;
  }

/*----------------------------------------------------------------------------
; From 6-card rank-only hand, return mask in low 15 bits that flags unique 
; discards by rank.  Assume hand sorted so that identical ranks are adjacent.
; Function not called by DLL but is included to show how Mask[] determined.
;---------------------------------------------------------------------------*/
int GetMask(char Hand[])
  {int n, m, t, Mask;

   /* loop through 15 possible discards */  
   for (Mask = t = n = 0; n < 5; n++)
      for (m = n+1; m < 6; (m++, t++))
         if (n == 0 || Hand[n] != Hand[n-1])
            if (m == n+1 || Hand[m] != Hand[m-1])
               Mask |= 1 << t;

   return Mask;
  }

/*----------------------------------------------------------------------------
; The 29 Mask values were computed as follows.  Any 6 input characters will
; work, as long as patterns match the 29 sections of LoadNdx.
;
; Mask[0]  = GetMask("654321");  // 0x7FFF
; Mask[1]  = GetMask("554321");  // 0x7E1F
; Mask[2]  = GetMask("544321");  // 0x71FD
; ...
; Mask[28] = GetMask("222111");  // 0x1005
;---------------------------------------------------------------------------*/
const int Mask[] =  /* 15-bit flags for non-repeat discards, see GetMask */
  { 0x7FFF,0x7E1F,0x71FD,0x4FBB,0x3B77,0x56EF,0x4E1B,0x3A17,0x560F,0x3175, 
    0x50ED,0x46AB,0x701D,0x41B9,0x0B33,0x1267,0x3015,0x500D,0x0A13,0x1207, 
    0x40A9,0x1065,0x4019,0x0131,0x0223,0x460B,0x4009,0x0203,0x1005
  };

/*----------------------------------------------------------------------------
; Set up to 15 global DsNdxTbl and HnNdxTbl elements. Low 15 mask bits flag
; non-repeat discards by rank.  Assume rank-only hand ordered high to low.
;---------------------------------------------------------------------------*/
void SetNdx(char Hand[], int *Ndx, int BitMask)
  {int n, m, t, i, j;
   char Cd[4];

   /* note: BitMask = GetMask(Hand);  */

   /* loop through non-repeat discards */
   for (t = n = 0; n < 5; n++)
      for (m = n+1; m < 6; (m++, t++))
         if (BitMask & 1 << t)
           {
            /* populate corresponding held cards using zero-based ranks */
            for (i = j = 0; i < 6; i++) 
               if (i != n && i != m) Cd[j++] = Hand[i] - 1;

            /* set discard index 0-90/hand index 0-1819 and bump Ndx */
            DsNdxTbl[(*Ndx)] = QkWts2[Hand[n] - 1] + Hand[m] - 1;
            HnNdxTbl[(*Ndx)++] = 
               QkWts0[Cd[0]] + QkWts1[Cd[1]] + QkWts2[Cd[2]] + Cd[3];
           }
  }

/*----------------------------------------------------------------------------
; Load global DsNdxTbl/HnNdxTbl with precomputed 0-90/0-1819 indexes.  Called 
; once over life of DLL.  These arrays speed up frequently called functions.
;---------------------------------------------------------------------------*/
void LoadNdx()
  {int u, v, w, x, y, z, Ndx;
   char Hn[6];
   
   /* set discard indexes for type 111111 hands */
   Ndx = 0;
   for (u = 13;  u > 5; u--)
   for (v = u-1; v > 4; v--)
   for (w = v-1; w > 3; w--)
   for (x = w-1; x > 2; x--)
   for (y = x-1; y > 1; y--)
   for (z = y-1; z > 0; z--)
     {
      Hn[0] = u; Hn[1] = v; Hn[2] = w; Hn[3] = x; Hn[4] = y; Hn[5] = z;
      SetNdx(Hn, &Ndx, Mask[0]);
     }

   /* set discard indexes for type 11112 hands */
   for (u = 13;  u > 4; u--)
   for (v = u-1; v > 3; v--)
   for (w = v-1; w > 2; w--)
   for (x = w-1; x > 1; x--)
   for (y = x-1; y > 0; y--)
     {
      /* 5 positions for the pair */
      Hn[0] = Hn[1] = u; Hn[2] = v; Hn[3] = w; Hn[4] = x; Hn[5] = y;/* 21111 */
      SetNdx(Hn, &Ndx, Mask[1]);

      Hn[1] = v;                                                    /* 12111 */
      SetNdx(Hn, &Ndx, Mask[2]);

      Hn[2] = w;                                                    /* 11211 */
      SetNdx(Hn, &Ndx, Mask[3]);

      Hn[3] = x;                                                    /* 11121 */
      SetNdx(Hn, &Ndx, Mask[4]);

      Hn[4] = y;                                                    /* 11112 */
      SetNdx(Hn, &Ndx, Mask[5]);
     }

   /* set discard indexes for type 1122 and 1113 hands */
   for (u = 13;  u > 3; u--)
   for (v = u-1; v > 2; v--)
   for (w = v-1; w > 1; w--)
   for (x = w-1; x > 0; x--)
     {
      /* 1122 hands--6 positions for the 2 pairs */
      Hn[0] = Hn[1] = u; Hn[2] = Hn[3] = v; Hn[4] = w; Hn[5] = x;   /* 2211 */
      SetNdx(Hn, &Ndx, Mask[6]);

      Hn[3] = w;                                                    /* 2121 */
      SetNdx(Hn, &Ndx, Mask[7]);

      Hn[4] = x;                                                    /* 2112 */
      SetNdx(Hn, &Ndx, Mask[8]);

      Hn[1] = v; Hn[4] = w;                                         /* 1221 */
      SetNdx(Hn, &Ndx, Mask[9]);

      Hn[4] = x;                                                    /* 1212 */
      SetNdx(Hn, &Ndx, Mask[10]);

      Hn[2] = w;                                                    /* 1122 */
      SetNdx(Hn, &Ndx, Mask[11]);

      /* 1113 hands--4 positions for the trips */
      Hn[0] = Hn[1] = Hn[2] = u; Hn[3] = v; Hn[4] = w; Hn[5] = x;   /* 3111 */
      SetNdx(Hn, &Ndx, Mask[12]);

      Hn[1] = Hn[2] = v;                                            /* 1311 */
      SetNdx(Hn, &Ndx, Mask[13]);

      Hn[2] = Hn[3] = w;                                            /* 1131 */
      SetNdx(Hn, &Ndx, Mask[14]);

      Hn[3] = Hn[4] = x;                                            /* 1113 */
      SetNdx(Hn, &Ndx, Mask[15]);
     }

   /* set discard indexes for type 123, 114, and 222 hands */
   for (u = 13;  u > 2; u--)
   for (v = u-1; v > 1; v--)
   for (w = v-1; w > 0; w--)
     {
      /* 123 hands--6 positions for pair/trips */
      Hn[0] = Hn[1] = Hn[2] = u; Hn[3] = Hn[4] = v; Hn[5] = w;      /* 321 */
      SetNdx(Hn, &Ndx, Mask[16]);

      Hn[4] = w;                                                    /* 312 */
      SetNdx(Hn, &Ndx, Mask[17]);

      Hn[2] = Hn[4] = v;                                            /* 231 */
      SetNdx(Hn, &Ndx, Mask[18]);

      Hn[3] = Hn[4] = w;                                            /* 213 */
      SetNdx(Hn, &Ndx, Mask[19]);

      Hn[1] = Hn[3] = v;                                            /* 132 */
      SetNdx(Hn, &Ndx, Mask[20]);

      Hn[3] = w;                                                    /* 123 */
      SetNdx(Hn, &Ndx, Mask[21]);

      /* 114 hands--3 positions for quad */
      Hn[0] = Hn[1] = Hn[2] = Hn[3] = u; Hn[4] = v; Hn[5] = w;      /* 411 */
      SetNdx(Hn, &Ndx, Mask[22]);

      Hn[1] = Hn[2] = Hn[3] = v;                                    /* 141 */
      SetNdx(Hn, &Ndx, Mask[23]);

      Hn[2] = Hn[3] = Hn[4] = w;                                    /* 114 */
      SetNdx(Hn, &Ndx, Mask[24]);

      /* 222 hands--1 position for 3 pairs */
      Hn[0] = Hn[1] = u; Hn[2] = Hn[3] = v; Hn[4] = Hn[5] = w;      /* 222 */
      SetNdx(Hn, &Ndx, Mask[25]);
     }

   /* set discard indexes for type 24 and 33 hands */
   for (u = 13;  u > 1; u--)
   for (v = u-1; v > 0; v--)
     {
      /* 24 hands--2 positions for pair/quad */
      Hn[0] = Hn[1] = Hn[2] = Hn[3] = u; Hn[4] = Hn[5] = v;         /* 42 */
      SetNdx(Hn, &Ndx, Mask[26]);

      Hn[2] = Hn[3] = v;                                            /* 24 */
      SetNdx(Hn, &Ndx, Mask[27]);

      /* 33 hands--1 position for 2 trips */
      Hn[0] = Hn[1] = Hn[2] = u; Hn[3] = Hn[4] = Hn[5] = v;         /* 33 */
      SetNdx(Hn, &Ndx, Mask[28]);
     }
  }

/****************************************************************************/
/************************* Risk Assessment Functions ************************/
/****************************************************************************/

/*----------------------------------------------------------------------------
; From dealer or pone score and peg point 0/1, return threshhold to go out.
; Score range is about -10 to 140, since peg routines use effective scores.
;---------------------------------------------------------------------------*/
int GetThresh(int Score, int PegPoint)
  {int ThreshHold;

   ThreshHold = 121 - Score - PegPoint;
   if (ThreshHold < 0)   ThreshHold = 0;
   if (ThreshHold > 121) ThreshHold = 121;

   return ThreshHold;
  }

/*----------------------------------------------------------------------------
; Inflate risk [-1.0,1,0] toward extremes using low threshhold (effective 
; points to go out for leader in game).  Expect threshhold in range [0,121].
; Function not currently used. 
;---------------------------------------------------------------------------*/
#define MIN_PWR 1.0
#define MAX_PWR 2.0

double BoostRisk(double Risk, int LowThresh)
  {double Power = 
     (MAX_PWR - MIN_PWR) * log(2.0 - LowThresh / 121.0) / log(2.0) + MIN_PWR;

   /* power increases from MIN_PWR toward MAX_PWR as game proceeds */
   if (Risk < 0.0) return pow(1.0 + Risk, Power) - 1.0;
   else            return 1.0 - pow(1.0 - Risk, Power);
  }

/*----------------------------------------------------------------------------
; From dealer threshhold, return risk assessment at deal based on shape.
; Range is -1.0 to 1.0 where low means push and high means balk.  Neutral is 
; zero.  If pone threshhold passed, risk assessment is negative of return
; value.  E.g., dealer on 4th street at 95 points has maximum push -1.0, while
; pone at same score has maximum balk 1.0.
;
; +1.0 balk                                      As game proceeds -->
;                   * *                          Min to Max: 10 points
;                  *    *                        Max to Min: 16 points
;                 *       *                                  ---------
; -*------95----*---105------*------121-->           Period: 26 points
;     *       *                 *       *                             
;       *    *                    *    *         Score: ..., 95, 105, 121
;         * *                       * *     Threshhold: ..., 26,  16,   0
; -1.0 push                                                           
---------------------------------------------------------------------------*/
#define PN_AVE  10              /* roughly pone total points per hand */
#define DL_AVE  16              /* roughly dealer total points per hand */
#define PERIOD (DL_AVE+PN_AVE)  /* points scored over two-hand cycle */
#define PHASE   0               /* shifts curve left/right (+/-) if not 0 */
#define PI      acos(-1.0)      /* 3.14159... */

double EvalShape(int ThreshHold)
  {int n = (ThreshHold - PHASE) % PERIOD, m = n - DL_AVE;

   /* value PI here forces double precision in cos() arguments */
   return (m < 0) ? -cos(n * PI / DL_AVE) : cos(m * PI / PN_AVE);
  }

/*----------------------------------------------------------------------------
; From dealer/pone (effective) scores at deal, return dealer risk in range 
; -1.0 to 1.0.  High value is caution.  Pone assessment is negative of return 
; value.  Threshholds below in range [0,121] and difference in range [-30,30].
;---------------------------------------------------------------------------*/
#define DL_EDGE    3     /* half of normal point swing for dealer on hand */
#define SKUNK_LIM  21    /* point difference to kick in skunk evaluation */
#define STREET     30    /* points in street */

double EvalRisk(int DlScr, int PnScr)
  {int DlThresh, PnThresh, LowThresh, Diff;
   double Risk;

   DlThresh = GetThresh(DlScr, 0);
   PnThresh = GetThresh(PnScr, 0);

   /* no loop unless skunk consideration kicks in */
   while (1)
     {Diff = PnThresh + DL_EDGE - DlThresh;

      if (Diff < 0)  /* pone ahead effectively */
        {LowThresh = PnThresh;
         if (-Diff <= SKUNK_LIM) break;
         if (DlThresh <= STREET) break;
         DlThresh -= STREET;
        }
      else           /* dealer ahead effectively */
        {LowThresh = DlThresh;
         if (Diff <= SKUNK_LIM)  break;
         if (PnThresh <= STREET) break;
         PnThresh -= STREET; 
        }
     }

   /* restore difference to range [-30,30] in case needed further */
   Diff -= DL_EDGE;

   /* combine dealer and pone shape assessments, keeping range [-1.0,1.0], */
   /* then weigh result linearly increasing as game proceeds  */
   Risk  = (EvalShape(DlThresh) + EvalShape(PnThresh)) / 2.0;
   Risk *=  1.0 - LowThresh / 121.0;

   return Risk;
  }

/*----------------------------------------------------------------------------
; Return risk -1.0 or +1.0 if end play pegging, else 0.0 if not near game end.
;---------------------------------------------------------------------------*/
double EndPlayRisk(int OwnPlayed, int OppPlayed, int OwnScr, int OppScr, 
                   int IsDealer, int HndScr)
  {
   /* account for guaranteed peg point */
   if (OwnPlayed >= OppPlayed) ++OppScr;
   else                        ++OwnScr;

   if (IsDealer)
     {OwnScr += 4 - OwnPlayed;  /* estimate gain of point per card left */
      if (OwnScr >= 121) return -1.0;  /* if might peg out, push */
      if (OppScr >= 106) return +1.0;  /* if opponent close, balk */
     }
   else
     {if (OppScr >= 121) return -1.0;  /* if opponent will go out, push */

      OwnScr += HndScr;         /* bump score by what hand will score */
      OppScr += 4 - OppPlayed;  /* estimate gain of point per card left */
      if (OwnScr >= 121 && OppScr < 121) return +1.0;   /* balk */
     }

   return 0.0;  /* indeterminate--not deemed in end play */
  }

/*----------------------------------------------------------------------------
; Return risk factor for opponent peg scoring plays in range [-1.0,1.0].
;---------------------------------------------------------------------------*/
double EvalPegRisk(char Cards[], int OwnPlayed, int OppPlayed, 
                   int OwnScr, int OppScr, int IsDealer)
  {int HndScr;
   char Card;
   double Risk;

   /* temporarily move cut to 5th position to score hand */
   Card = Cards[4]; 
   Cards[4] = Cards[6];
   HndScr = Score5(Cards, 0, 0);  /* not crib, not parse */
   Cards[4] = Card;

   /* early risk determination from end game check (-1.0, 0.0, or 1.0) */
   Risk = EndPlayRisk(OwnPlayed, OppPlayed, OwnScr, OppScr, IsDealer, HndScr);
   
   /* if early check indeterminate */
   if (Risk == 0.0)
     {
      /* adjust own score per average 8-point hand--may be negative now */
      OwnScr += HndScr - 8;

      /* back up scores slightly, estimating point per played card */
      OwnScr -= OwnPlayed;
      OppScr -= OppPlayed;

      /* get risk -1.0 to 1.0 based on effective scores */
      if (IsDealer) Risk =  EvalRisk(OwnScr, OppScr);
      else          Risk = -EvalRisk(OppScr, OwnScr);
     }
   
   return Risk;
  }

/*----------------------------------------------------------------------------
; Based on risk, return scale factor, usually 1.0, for opponent scoring  
; during pegging assessment.  Risk +1.0 or -1.0 (from EndPlayRisk) is handled
; specially below.
;---------------------------------------------------------------------------*/
#define PEG_SPEC    3.5   /* factor for special handling */
#define PEG_FAC     1.0   /* factor to adjust impact */
#define PEG_CUT_OFF 0.5   /* leave default scale unless near extremes */

double GetPegScale(double Risk)
  {double Scale = 1.0;  /* default scale factor */

   /* most of time, scale will be neutral 1.0, but when risk near -1.0 or 1.0,
   /* opponent scale factor set to lower or higher value */
   if      (Risk == -1.0)         Scale = 1.0 / (1.0 + PEG_SPEC);
   else if (Risk <= -PEG_CUT_OFF) Scale = 1.0 / (1.0 - PEG_FAC * Risk);

   if      (Risk ==  1.0)         Scale = 1.0 + PEG_SPEC;
   else if (Risk >=  PEG_CUT_OFF) Scale = 1.0 + PEG_FAC * Risk;

   return Scale;
  }

/*----------------------------------------------------------------------------
; Based on risk, return scale factor, usually 1.0, for opponent crib scoring
; during discard assessment.  DIS_FAC 0.0 forces 1.0 scale.
;---------------------------------------------------------------------------*/
#define DIS_FAC     0.1   /* factor to adjust impact */
#define DIS_CUT_OFF 0.8   /* leave default scale unless near extremes */

double GetDisScale(double Risk)
  {double Scale = 1.0;  /* default scale factor */

   /* most of time, scale will be neutral 1.0, but when risk near -1.0 or 1.0,
   /* opponent scale factor set to lower or higher value */
   if (Risk <= -DIS_CUT_OFF) Scale = 1.0 / (1.0 - DIS_FAC * Risk);
   if (Risk >=  DIS_CUT_OFF) Scale = 1.0 + DIS_FAC * Risk;

   return Scale;
  }

/****************************************************************************/
/***************************** Discard Functions ****************************/
/****************************************************************************/

/*----------------------------------------------------------------------------
; Return 5-card hand score (crib if flagged).  Zero suit voids flushes/nobs
; and zero rank excluded from 15s/runs/pairs.  This allows dummy zero cards
; for scoring shorter hands.  Cut is assumed to be 5th card.  If garbage
; bytes passed, no harm but output score meaningless.  If parse flagged,
; return 14 bits of parse info instead for VB text scoring display.
;---------------------------------------------------------------------------*/
int Score5(char Hand[], int IsCrib, int IsParse)
  {int  n, m, Sum, Tot = 0, Bits, Save, Cnt[16];
   char Jack;

   /* check flush--zero suit voids flush */
   Sum = SuitOf(Hand[0]) & Hand[1] & Hand[2] & Hand[3];
   if (Sum) Tot = 4 + (Sum & Hand[4] ? 1: 0);
   if (Tot == 4 && IsCrib) Tot = 0;

   /* check nobs--zero suit voids nobs */
   m = SuitOf(Hand[4]); Jack = m | 11;
   if (m && (Hand[0] == Jack || Hand[1] == Jack ||
             Hand[2] == Jack || Hand[3] == Jack)) Tot += 1;

   if (IsParse)
     {Bits = (Tot == 1 || Tot == 6) ? 1 : 0;  /* 1 bit for nobs */
      Bits = Bits << 1 | (Tot > 1 ? 1 : 0);   /* 1 bit for flush */
      Bits = Bits << 1 | (Tot > 4 ? 1 : 0);   /* 1 bit for 5-flush */
      Save = Tot;
     }

   /* check 15-counts--zero rank ignored as dummy 15-card */
   Sum = 0;
   for (n = 0; n < 5; n++)
      Sum += Cnt[n] = (m = CountOf(RankOf(Hand[n]))) ? m : 15;

   if (Sum == 15) Tot += 2;
   for (n = 0; n < 5; n++)
     {if (Sum - Cnt[n] == 15) Tot += 2;
      for (m = n+1; m < 5; m++)
        {if (Cnt[n] + Cnt[m] == 15) Tot += 2;
         if (Sum - Cnt[n] - Cnt[m] == 15) Tot += 2;
        }
      }

   if (IsParse)
     {Bits = Bits << 4 | (Tot - Save) >> 1;  /* 4 bits for count of 15's */
      Save = Tot;
     }

   /* check pairs/trips/quads--zero rank ignored */
   for (n = 0; n < 16; n++) Cnt[n] = 0;
   for (n = 0; n < 5; n++)
      if (m = RankOf(Hand[n])) Tot += Cnt[m]++ << 1;

   if (IsParse)
     {Bits = Bits << 3 | (Tot - Save) >> 1;  /* 3 bits for count of pairs */
      Save = Tot;
     }

   /* check runs--zero rank ignored */
   Sum = 0; m = 1;
   for (n = 1; n < 14; n++)
     {if (Cnt[n]) (++Sum, m *= Cnt[n]);
      else if (Sum < 3) (Sum = 0, m = 1);
      else break;
     }
   if (Sum >= 3) Tot += Sum * m;

   if (IsParse)
      Bits = Bits << 4 | Tot - Save;       /* 4 bits for run points */

   return IsParse ? Bits : Tot;
  }

/*----------------------------------------------------------------------------
; Return defensive spread of 4-card hand (maximum is 6*4 = 24).
;---------------------------------------------------------------------------*/
int Spread(char Hand[])
  {int n, m, Diff, Tot = 0;

   /* sum absolute value of pairwise rank differences, each capped at 4 */
   for (n = 0; n < 3; n++) for (m = n+1; m < 4; m++)
     {Diff = RankOf(Hand[n]) - RankOf(Hand[m]);
      Diff = Diff < 0 ? -Diff : Diff;
      Tot += Diff > 4 ? 4 : Diff;
     }

   return Tot;
  }

/*----------------------------------------------------------------------------
; Bump PnFreq/DlFreq arrays by scaled weights for discards.  This is core
; routine in discard section, so three different speed optimizations are used.
;
; First, the index pointer array has been precomputed.  Second, the short loop
; has been unrolled as a switch statement.  Third, an incremental weight
; difference (from prior LoadTot() call) rather than an absolute weight is
; used, since the weight difference is frequently zero.
;---------------------------------------------------------------------------*/
void BumpTot(int Wt, int PrevNdx, int MaxWt, int Len, int *Ndx, 
             double PnFreq[], double DlFreq[], short WtPrev[])
  {double Scale;
   short *PnPtr, *DlPtr;
   char  *NdxPtr;

   /* bump pone and dealer totals by scaled weight difference, if any */
   if (Wt -= WtPrev[PrevNdx])
     {
      Scale = (double)Wt / MaxWt;

      NdxPtr = DsNdxTbl + *Ndx;
      PnPtr  = PnTbl    + *Ndx;
      DlPtr  = DlTbl    + *Ndx;

      /* following is unrolled loop */
      switch (Len)
        {
         case 15:
            if (PnPtr[14]) PnFreq[(NdxPtr[14])] += PnPtr[14] * Scale;
            if (DlPtr[14]) DlFreq[(NdxPtr[14])] += DlPtr[14] * Scale;
            if (PnPtr[13]) PnFreq[(NdxPtr[13])] += PnPtr[13] * Scale;
            if (DlPtr[13]) DlFreq[(NdxPtr[13])] += DlPtr[13] * Scale;
            if (PnPtr[12]) PnFreq[(NdxPtr[12])] += PnPtr[12] * Scale;
            if (DlPtr[12]) DlFreq[(NdxPtr[12])] += DlPtr[12] * Scale;
            if (PnPtr[11]) PnFreq[(NdxPtr[11])] += PnPtr[11] * Scale;
            if (DlPtr[11]) DlFreq[(NdxPtr[11])] += DlPtr[11] * Scale;
         case 11:
            if (PnPtr[10]) PnFreq[(NdxPtr[10])] += PnPtr[10] * Scale;
            if (DlPtr[10]) DlFreq[(NdxPtr[10])] += DlPtr[10] * Scale;
            if (PnPtr[ 9]) PnFreq[(NdxPtr[ 9])] += PnPtr[ 9] * Scale;
            if (DlPtr[ 9]) DlFreq[(NdxPtr[ 9])] += DlPtr[ 9] * Scale;
            if (PnPtr[ 8]) PnFreq[(NdxPtr[ 8])] += PnPtr[ 8] * Scale;
            if (DlPtr[ 8]) DlFreq[(NdxPtr[ 8])] += DlPtr[ 8] * Scale;
         case 8:
            if (PnPtr[ 7]) PnFreq[(NdxPtr[ 7])] += PnPtr[ 7] * Scale;
            if (DlPtr[ 7]) DlFreq[(NdxPtr[ 7])] += DlPtr[ 7] * Scale;
         case 7:
            if (PnPtr[ 6]) PnFreq[(NdxPtr[ 6])] += PnPtr[ 6] * Scale;
            if (DlPtr[ 6]) DlFreq[(NdxPtr[ 6])] += DlPtr[ 6] * Scale;
         case 6:
            if (PnPtr[ 5]) PnFreq[(NdxPtr[ 5])] += PnPtr[ 5] * Scale;
            if (DlPtr[ 5]) DlFreq[(NdxPtr[ 5])] += DlPtr[ 5] * Scale;
         case 5:
            if (PnPtr[ 4]) PnFreq[(NdxPtr[ 4])] += PnPtr[ 4] * Scale;
            if (DlPtr[ 4]) DlFreq[(NdxPtr[ 4])] += DlPtr[ 4] * Scale;
         case 4:
            if (PnPtr[ 3]) PnFreq[(NdxPtr[ 3])] += PnPtr[ 3] * Scale;
            if (DlPtr[ 3]) DlFreq[(NdxPtr[ 3])] += DlPtr[ 3] * Scale;
         case 3:
            if (PnPtr[ 2]) PnFreq[(NdxPtr[ 2])] += PnPtr[ 2] * Scale;
            if (DlPtr[ 2]) DlFreq[(NdxPtr[ 2])] += DlPtr[ 2] * Scale;
            if (PnPtr[ 1]) PnFreq[(NdxPtr[ 1])] += PnPtr[ 1] * Scale;
            if (DlPtr[ 1]) DlFreq[(NdxPtr[ 1])] += DlPtr[ 1] * Scale;
            if (PnPtr[ 0]) PnFreq[(NdxPtr[ 0])] += PnPtr[ 0] * Scale;
            if (DlPtr[ 0]) DlFreq[(NdxPtr[ 0])] += DlPtr[ 0] * Scale;
            break;

         default: ;
            /* does not occur--semi-colon needed else omit default */
        }
     }

   /* input weight (changed above) is saved to array by summing here */
   WtPrev[PrevNdx] += Wt;
   *Ndx += Len;
  }

/*----------------------------------------------------------------------------
; Load PnFreq/DlFreq arrays with weights for discards to all 6-card hands out
; of 52-card deck with 7 cards missing.  May specify incremental change from
; prior call, provided an initial call passed 0 as flag.  These on-the-fly
; frequency tables, which vary depending on the missing cards, replace static 
; average tables used in the original Cribbage '97.
;---------------------------------------------------------------------------*/
void LoadTot(char Hand[], int IncrementalLoad, 
             double PnFreq[], double DlFreq[], short WtPrev[])
  {int u, v, w, x, y, z, n, Wt, Ndx, WtRow[16];

   /* undo prior normalization if incremental load, else initialize arrays */
   if (IncrementalLoad)
      for (n = 0; n < 91; n++) (PnFreq[n] *= SUM45, DlFreq[n] *= SUM45);
   else
     {
      for (n = 0; n < 91; n++) PnFreq[n] = DlFreq[n] = 0.0;
      for (n = 0; n < RNKSIZE; n++) WtPrev[n] = 0;
     }

   /* initialize rank weight counts - input Hand[] excludes suits */
   for (n = 1; n < 14; n++) WtRow[n] = 21;
   for (n = 0; n < 7; n++) if (WtRow[Hand[n]] > 1) WtRow[Hand[n]] -= 5;

   /* update sum arrays for type 111111 hands */
   n = Ndx = 0;
   for (u = 13;  u > 5; u--)
   for (v = u-1; v > 4; v--)
   for (w = v-1; w > 3; w--)
   for (x = w-1; x > 2; x--)
   for (y = x-1; y > 1; y--)
   for (z = y-1; z > 0; z--)
     {
      Wt = Wgts[WtRow[u]] * Wgts[WtRow[v]] * Wgts[WtRow[w]] *
           Wgts[WtRow[x]] * Wgts[WtRow[y]] * Wgts[WtRow[z]];        /* 111111 */
      BumpTot(Wt, n++, 4096, 15, &Ndx, PnFreq, DlFreq, WtPrev);
     }

   /* update sum arrays for type 11112 hands */
   for (u = 13;  u > 4; u--)
   for (v = u-1; v > 3; v--)
   for (w = v-1; w > 2; w--)
   for (x = w-1; x > 1; x--)
   for (y = x-1; y > 0; y--)
     {
      /* 5 positions for the pair */
      Wt = Wgts[1+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[  WtRow[x]] * Wgts[  WtRow[y]];                     /* 21111 */
      BumpTot(Wt, n++, 1536, 11, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[  WtRow[x]] * Wgts[WtRow[y]];                       /* 12111 */
      BumpTot(Wt, n++, 1536, 11, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[1+WtRow[w]] *
           Wgts[  WtRow[x]] * Wgts[  WtRow[y]];                     /* 11211 */
      BumpTot(Wt, n++, 1536, 11, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[1+WtRow[x]] * Wgts[  WtRow[y]];                     /* 11121 */
      BumpTot(Wt, n++, 1536, 11, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[  WtRow[x]] * Wgts[1+WtRow[y]];                     /* 11112 */
      BumpTot(Wt, n++, 1536, 11, &Ndx, PnFreq, DlFreq, WtPrev);
     }

   /* update sum arrays for type 1122 and 1113 hands */
   for (u = 13;  u > 3; u--)
   for (v = u-1; v > 2; v--)
   for (w = v-1; w > 1; w--)
   for (x = w-1; x > 0; x--)
     {
      /* 1122 hands--6 positions for the 2 pairs */
      Wt = Wgts[1+WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[  WtRow[x]];                                        /* 2211 */
      BumpTot(Wt, n++, 576, 8, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[1+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[1+WtRow[w]] *
           Wgts[  WtRow[x]];                                        /* 2121 */
      BumpTot(Wt, n++, 576, 8, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[1+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[1+WtRow[x]];                                        /* 2112 */
      BumpTot(Wt, n++, 576, 8, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[1+WtRow[w]] *
           Wgts[  WtRow[x]];                                        /* 1221 */
      BumpTot(Wt, n++, 576, 8, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[1+WtRow[x]];                                        /* 1212 */
      BumpTot(Wt, n++, 576, 8, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[1+WtRow[w]] *
           Wgts[1+WtRow[x]];                                        /* 1122 */
      BumpTot(Wt, n++, 576, 8, &Ndx, PnFreq, DlFreq, WtPrev);

      /* 1113 hands--4 positions for the trips */
      Wt = Wgts[2+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[  WtRow[x]];                                        /* 3111 */
      BumpTot(Wt, n++, 256, 7, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[2+WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[  WtRow[x]];                                        /* 1311 */
      BumpTot(Wt, n++, 256, 7, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[2+WtRow[w]] *
           Wgts[  WtRow[x]];                                        /* 1131 */
      BumpTot(Wt, n++, 256, 7, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[2+WtRow[x]];                                        /* 1113 */
      BumpTot(Wt, n++, 256, 7, &Ndx, PnFreq, DlFreq, WtPrev);
     }

   /* update sum arrays for type 123, 114, and 222 hands */
   for (u = 13;  u > 2; u--)
   for (v = u-1; v > 1; v--)
   for (w = v-1; w > 0; w--)
     {
      /* 123 hands--6 positions for pair/trips */
      Wt = Wgts[2+WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[  WtRow[w]];  /* 321 */
      BumpTot(Wt, n++, 96, 5, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[2+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[1+WtRow[w]];  /* 312 */
      BumpTot(Wt, n++, 96, 5, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[1+WtRow[u]] * Wgts[2+WtRow[v]] * Wgts[  WtRow[w]];  /* 231 */
      BumpTot(Wt, n++, 96, 5, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[1+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[2+WtRow[w]];  /* 213 */
      BumpTot(Wt, n++, 96, 5, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[2+WtRow[v]] * Wgts[1+WtRow[w]];  /* 132 */
      BumpTot(Wt, n++, 96, 5, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[2+WtRow[w]];  /* 123 */
      BumpTot(Wt, n++, 96, 5, &Ndx, PnFreq, DlFreq, WtPrev);

      /* 114 hands--3 positions for quad */
      Wt = Wgts[3+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]];  /* 411 */
      BumpTot(Wt, n++, 16, 4, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[3+WtRow[v]] * Wgts[  WtRow[w]];  /* 141 */
      BumpTot(Wt, n++, 16, 4, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[3+WtRow[w]];  /* 114 */
      BumpTot(Wt, n++, 16, 4, &Ndx, PnFreq, DlFreq, WtPrev);

      /* 222 hands--1 position for 3 pairs */
      Wt = Wgts[1+WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[1+WtRow[w]];  /* 222 */
      BumpTot(Wt, n++, 216, 6, &Ndx, PnFreq, DlFreq, WtPrev);
     }

   /* update sum arrays for type 24 and 33 hands */
   for (u = 13;  u > 1; u--)
   for (v = u-1; v > 0; v--)
     {
      /* 24 hands--2 positions for pair/quad */
      Wt = Wgts[3+WtRow[u]] * Wgts[1+WtRow[v]];                     /* 42 */
      BumpTot(Wt, n++, 6, 3, &Ndx, PnFreq, DlFreq, WtPrev);

      Wt = Wgts[1+WtRow[u]] * Wgts[3+WtRow[v]];                     /* 24 */
      BumpTot(Wt, n++, 6, 3, &Ndx, PnFreq, DlFreq, WtPrev);

      /* 33 hands--1 position for 2 trips */
      Wt = Wgts[2+WtRow[u]] * Wgts[2+WtRow[v]];                     /* 33 */
      BumpTot(Wt, n++, 16, 3, &Ndx, PnFreq, DlFreq, WtPrev);
     }

   /* normalize sum arrays to frequencies */
   for (n = 0; n < 91; n++) (PnFreq[n] /= SUM45, DlFreq[n] /= SUM45);
  }

/*----------------------------------------------------------------------------
; From 6-card hand, return 15 average dealer/pone crib scores in output arrays.
; This routine adapted from DS.C code, which in turn was based on GEN.C output.
;
; Main loop computes rank-based averages based on arrays PnTbl/DlTbl output
; from GEN.C program.  These arrays each sum to (52 6) = 20358520, which is
; total number of possible hands.  The array entries give weights for the
; possible discards by rank from all 18395 possible rank-type hands.

; E.g., for 111111 rank-type hands (no pairs), there are 15 array entries --
; one for each possible discard by rank.  The total weight for the 15 entries
; is 4096.  These are total possible hands within the 111111 rank-type due to
; suit combinations.  Usually, one entry dominates with a large weight, but
; several others will have smaller weights due to flushes and nobs.
;
; The main loop here effectively reduces its view of the two tables to 13
; subsets of (45 6) = 8145060 hands each.  These correspond to card decks with
; the 6 known hand cards plus tentative cut removed.  From these subset views,
; 91-entry opponent discard frequency tables are generated.  From these
; frequency tables, the expected crib averages are computed for each possible
; discard by rank from the 6-card hand.
;---------------------------------------------------------------------------*/
void SetAve2(char Hand[], double PnAve[], double DlAve[])
  {int n, m, t, Cut, i, j, k, QtWt, CutWt, IncrementalLoad, Ct[16], Cnt, Sum;
   char Hn[7];
   double DlSum, PnSum, PnJack, DlJack, Frac, PnFreq[91], DlFreq[91];
   short WtPrev[RNKSIZE];

   /* clear entries where results will accumulate */
   PnJack = DlJack = 0.0;
   for (t = 0; t < 15; t++) PnAve[t] = DlAve[t] = 0.0;

   /* accumulate rank-based averages first */
   IncrementalLoad = 0;
   for (Cut = 1; Cut < 14; Cut++)
     {
      CutWt = 4; Hn[6] = Cut;
      for (k = 0; k < 6; k++) if (Cut == (Hn[k] = RankOf(Hand[k]))) --CutWt;

      if (CutWt > 0)
        {
         /* first load is not incremental, rest are incremental */
         LoadTot(Hn, IncrementalLoad++, PnFreq, DlFreq, WtPrev);

         /* loop through 15 possible discards */
         Hn[4] = Cut;
         for (t = n = 0; n < 5; n++)
         for (m = n+1; m < 6; (m++, t++))
           {
            Hn[2] = RankOf(Hand[n]); Hn[3] = RankOf(Hand[m]);
            DlSum = PnSum = 0.0;
            
            /* loop through 91 possible opponent discards - for impossible */
            /* opponent discards, frequencies from LoadTot will be zero */
            for ((k = 0, i = 1); (Hn[0] = i) < 14; i++)
            for (j = 1; (Hn[1] = j) <= i; (j++, k++))
              {
               /* no suit info here - nobs/flush handled after Cut loop */
               QtWt  = Score5(Hn, 1, 0);

               /* sums here reflect opponent frequency */
               PnSum += DlFreq[k] * QtWt;
               DlSum += PnFreq[k] * QtWt;
              }

            PnAve[t] += PnSum * CutWt;
            DlAve[t] += DlSum * CutWt;
           }

         /* accumulate opponent jack frequency when jack is not cut */
         if (Cut != 11)
         for ((k = 0, i = 1); i < 14; i++)
         for (j = 1; j <= i; (j++, k++))
         if (t = CutWt * ((i == 11) + (j == 11)))
           {
            PnJack += DlFreq[k] * t;
            DlJack += PnFreq[k] * t;
           }
        }
     }

   /* normalize results over the 46 cuts */
   PnJack /= 46; DlJack /= 46;
   for (t = 0; t < 15; t++) (PnAve[t] /= 46, DlAve[t] /= 46);

   /* remaining code makes nobs/flush adjustments with suit info */

   /* get hand suit distribution - longer array helps prevent bug */
   /* overruns during testing, e.g., bad suit in high nibble */
   Ct[1] = Ct[2] = Ct[4] = Ct[8] = 0;
   for (n = 0; n < 6; n++) ++Ct[SuitOf(Hand[n]) >> 4];

   /* get jack count and jack split sum */
   for (Cnt = Sum = n = 0; n < 6; n++)
      if (RankOf(Hand[n]) == 11)
         (++Cnt, Sum += Ct[SuitOf(Hand[n]) >> 4]);

   /* from jack count/split sum, get suit-sensitive fraction near 0.25 */
   Frac = k = 42 + Sum - 12 * Cnt;
   if (k) Frac /= (42 + Cnt) * (4 - Cnt);

   /* adjust opponent nobs by suit-sensitive fraction */
   PnJack *= Frac; DlJack *= Frac;

   /* update averages */
   for (t = n = 0; n < 5; n++)
   for (m = n+1; m < 6; (m++, t++))
     {
      /* opponent discard contribution to nobs */
      PnAve[t] += PnJack;
      DlAve[t] += DlJack;

      k = 0;
      if (RankOf(Hand[n]) == 11) k += 13 - Ct[SuitOf(Hand[n]) >> 4];
      if (RankOf(Hand[m]) == 11) k += 13 - Ct[SuitOf(Hand[m]) >> 4];
      Frac = k / 46.0;  /* need cast or decimal here */

      /* own discard contribution to nobs */
      PnAve[t] += Frac;
      DlAve[t] += Frac;

      /* small contribution for possible crib flush */
      if (SuitOf(Hand[n]) == SuitOf(Hand[m]))
        {
         /* get expected points if missing 3 cards were random */
         k = 13 - Ct[SuitOf(Hand[n]) >> 4];
         Frac = 5 * (k * (k - 1) * (k - 2)) / 6;    /* 5 * (k 3) */
         Frac /= 15180;                             /*    (46 3) */

         /* reduce slightly per GEN.C estimates for opponent throws */
         PnAve[t] += Frac * DL_SAMESUITREDUCTION;
         DlAve[t] += Frac * PN_SAMESUITREDUCTION;
        }
     }
  }

/*----------------------------------------------------------------------------
; From 15 possible discards for 6-card hand, fill output arrays with minimum/
; maximum/average hand scores, defensive spread, and minimum crib scores.
;---------------------------------------------------------------------------*/
void SetAve4(char Hand[], double HnAve[], int HnMn[], int HnMx[], 
             int Sprd[], int CbMin[])
  {int n, m, k, t, Rank, Suit, Tot, Mn, Mx;
   char Hn[6], Cut;

   /* loop through 15 possible discards */
   for (t = n = 0; n < 5; n++)
   for (m = n+1; m < 6; (m++, t++))
     { 
      /* copy hand, then swap discards to end positions */
      for (k = 0; k < 6; k++) Hn[k] = Hand[k];
      k = Hn[m]; Hn[m] = Hn[5]; Hn[5] = k;
      k = Hn[n]; Hn[n] = Hn[4]; Hn[4] = k;

      /* set crib minimum at 0 or 2 */
      Mn = RankOf(Hn[4]); Mx = RankOf(Hn[5]); k = CountOf(Mn) + CountOf(Mx);
      CbMin[t] = (k == 15 || k == 5 || Mn == 5 || Mx == 5 || Mn == Mx) ? 2 : 0;
     
      /* evaluate 52 cuts, ignoring 6 matches to input cards */
      Tot = 0; Mn = 999; Mx = 0;
      for (Suit = 0x10; Suit < 0x100; Suit <<= 1)
         for (Rank = 13; Rank; Rank--)
            if (Hn[4] = Cut = Rank | Suit,
                  Cut != Hand[0] && Cut != Hand[1] && Cut != Hand[2] &&
                  Cut != Hand[3] && Cut != Hand[4] && Cut != Hand[5])
            {
             Tot += (k = Score5(Hn, 0, 0));  /* not crib, not parse */
             if (k < Mn) Mn = k;
             if (k > Mx) Mx = k;
            }

      /* set hand average/minimum/maximum/spread */
      HnAve[t] = Tot / 46.0;
      HnMn[t]  = Mn;
      HnMx[t]  = Mx;
      Sprd[t]  = Spread(Hn);
     }
  }

/*----------------------------------------------------------------------------
; Return double rounded to int away from zero in positive/negative direction.
;---------------------------------------------------------------------------*/
int Round(double Val)
  {
   return (int)(Val > 0.0 ? Val + 0.5 : Val - 0.5);
  }

/*----------------------------------------------------------------------------
; From dealer/pone scores and 6-card hand, return pone discards in low byte
; and dealer discards in high byte.  Discards are indexes 0-5.  Hand maxima
; are computed below, but not used.  Risk assessment is barely used.
;---------------------------------------------------------------------------*/
int PickDis(char Hand[], int DlScr, int PnScr)
  {int n, m, t, DlPicks, PnPicks, DlThresh, PnThresh, PnMxSpr, DlMxSpr,
       HnMn[15], HnMx[15], Sprd[15], CbMin[15], AdjHn, AdjPn;
   double PnScale, PnMx, DlMx, PnAvMx, PnComb, DlComb, 
       PnAve[15], DlAve[15], HnAve[15]; 

   /* get threshholds (points to go out) */
   DlThresh = GetThresh(DlScr, 1);
   PnThresh = GetThresh(PnScr, 0);

   /* get hand averages/minimums/maximums/spreads, crib minimums, */ 
   /* and pone/dealer crib averages for the 15 possible discards */
   SetAve4(Hand, HnAve, HnMn, HnMx, Sprd, CbMin);
   SetAve2(Hand, PnAve, DlAve);

   /* loop through 15 possible discard picks--final PnMx may be negative */
   DlMxSpr = PnMxSpr = -1;
   PnAvMx = -1.0;
   PnMx = DlMx = -999.0;  /* small enough */
   for (t = n = 0; n < 5; n++) 
   for (m = n+1; m < 6; (m++, t++))
     {
      /* get effective pone hand/crib score adjustments, using GEN.C overall */
      /* averages for pone/crib hands */
      AdjHn = Round(HnAve[t] - PN_HANDAVERAGE);
      AdjPn = Round(PnAve[t] - CB_HANDAVERAGE);
       
      /* get crib scale factor for pone based on risk--usually 1.0 */
      PnScale = GetDisScale(-EvalRisk(DlScr + AdjPn, PnScr + AdjHn));

      /* default handling--ignore risk assessment for dealer */
      DlComb = HnAve[t] + DlAve[t];
      PnComb = HnAve[t] - PnScale * PnAve[t];

      /* compare miminum scores to threshholds for endgame handling */
      if (HnMn[t] + CbMin[t] >= DlThresh && Sprd[t] > DlMxSpr)
         (DlMxSpr = Sprd[t], DlPicks = n | m << 4);
      if (HnMn[t] >= PnThresh && Sprd[t] > PnMxSpr)
         (PnMxSpr = Sprd[t], PnPicks = n | m << 4);

      /* compare pone average scores to threshholds for endgame handling */
      if (PnMxSpr < 0 && HnAve[t] >= PnThresh && HnAve[t] > PnAvMx)
         (PnAvMx = HnAve[t], PnPicks = n | m << 4);

      /* normal handling otherwise--spread maxes are -1 initially */
      if (DlMxSpr < 0 && DlComb > DlMx)
         (DlMx = DlComb, DlPicks = n | m << 4);
      if (PnMxSpr < 0 && PnAvMx < 0.0 && PnComb > PnMx)
         (PnMx = PnComb, PnPicks = n | m << 4);
 
     }
   
   /* pack indexes into return value - format expected by export function */
   return DlPicks << 8 | PnPicks;
  }   

/****************************************************************************/
/***************************** Pegging Functions ****************************/
/****************************************************************************/

/*----------------------------------------------------------------------------
; From played cards/length since last reset, return peg score (ignore go).
;---------------------------------------------------------------------------*/
int ScorePeg(char Cards[], int Length)
  {int Score = 0, Count = 0, Rank, n, Cnt[16];

   /* eliminate simple cases */
   if (Length < 2) return Score;

   /* clear multiplicity array */
   for (n = 1; n < 15; n++) Cnt[n] = 0;

   /* get count and set multiplicities */
   for (n = 0; n < Length; n++)
     {Rank = RankOf(Cards[n]);
      ++Cnt[Rank];
      Count += CountOf(Rank);
     }

   /* handle 15/31 count */
   if (Count == 15 || Count == 31) Score += 2;

   /* look for pair/trip/quad at end */
   Count = 0;
   while (n && Rank == RankOf(Cards[--n])) ++Count;
   Count *= Count - 1;
   if (Count) return Score + Count;

   /* look for run, beginning with largest possible */
   for (n = 0; n <= Length - 3; --Cnt[RankOf(Cards[n++])])
     {Count = Rank = 1;
      while (Cnt[Rank++] == 0);
      while (Cnt[Rank++]) ++Count;
      if (Count == Length - n) return Score + Count;
     }

   return Score;
  }

/*----------------------------------------------------------------------------
; Preliminary check to insure legal pegging reconstruction.  Assume 7-11 legal
; unique cards and 0-4 left to play for player.  See similar BuildPl().
;---------------------------------------------------------------------------*/
int ChkPeg(char Cards[], int Length, int CardsLeft, int IsOwnTurn, int IsRestart)
  {int n, PegCount = 0, GoCnt = 0;
   HANDSTRUCT Own, Opp;
   LPHANDSTRUCT Cur;

   Own.Hand = Cards;   Own.Len = 4-CardsLeft; Own.Ndx = 0;
   Opp.Hand = Cards+7; Opp.Len = Length-7;    Opp.Ndx = 0;

   /* replay pegging to own turn, aborting on any inconsistency */
   while (Own.Ndx < Own.Len || Opp.Ndx < Opp.Len || !IsOwnTurn)
     {Cur = IsOwnTurn ? &Own : &Opp;

      if (Cur->Ndx >= Cur->Len) n = 32;
      else n = PegCount + CountOf(RankOf(Cur->Hand[Cur->Ndx]));

      if (n > 31)
         /* insure 'go' not mistake, i.e., check if later card is */
         /* playable--unknown opponent cards are at worst face cards */
        {for (n = Cur->Ndx+1; n < (IsOwnTurn ? 4 : Cur->Len); n++)
            if (PegCount + CountOf(RankOf(Cur->Hand[n])) <= 31) return 1;

         if (!IsOwnTurn && Cur->Len < 4 && PegCount <= 21) return 1;

         ++GoCnt;
        }
      else
        {PegCount = n; GoCnt = 0; Cur->Ndx++;
        }

      if (PegCount == 31 || GoCnt > 1) PegCount = GoCnt = 0;


      IsOwnTurn = !IsOwnTurn;
     }

   /* if restarting and count non-zero, insure legal restart */
   if (IsRestart && PegCount)
     {if (GoCnt && Opp.Ndx < Opp.Len) return 1;
      for (n = Own.Ndx; n < 4; n++)
         if (PegCount + CountOf(RankOf(Own.Hand[n])) <= 31) return 1;
     }

   return 0;
  }

/*----------------------------------------------------------------------------
; From own/opponent scores, own/opponent cards played, dealer/restart flags,
; and known cards (own played cards in order/unplayed cards/discards, then cut
; card, then opponent played cards in order), return zero if go, else return
; index 1-4 of chosen play.
;---------------------------------------------------------------------------*/
int ChoosePeg(char Cards[], int OwnPlayed, int OppPlayed, int OwnScr, 
              int OppScr, int IsDealer, int IsRestart)
  {int n;
   double HnFreq[1820];  /* pone or dealer 4-card frequencies by rank */
   int    CtAry[16];     /* rank count array - only entries 1-13 referenced */
   double WtAry[16];     /* rank weight array - only entries 1-13 referenced */
   char   PlAry[8];      /* played card array */
   PEGSTRUCT PegData;
 
   PegData.CtPtr  = CtAry;
   PegData.WtPtr  = WtAry;
   PegData.PlPtr  = PlAry;              /* advanced by Replay() */
   PegData.HnPtr  = Cards; 
   PegData.GoFlgs = 0;                  /* set by Replay() */
   PegData.OppMin = 0;                  /* set by Replay() */
   PegData.HnFlgs = 0x0F >> OwnPlayed;  /* reverse order */
   PegData.OppLft = 4 - OppPlayed;
   PegData.PegCnt = 0;                  /* set by Replay() */
   PegData.PlCnt  = 0;                  /* set by Replay() */
   PegData.Depth  = 1;                  /* seems best */
   PegData.Risk   = 0.0;                /* set below */
   PegData.Bits   = 0;                  /* set by PickPeg() */

   /* reconstruct pegging to current state */
   BuildCt(Cards, 7+OppPlayed, CtAry);
   BuildPl(Cards, OwnPlayed, OppPlayed, !IsDealer, PlAry);
   Replay(&PegData, OwnPlayed, OppPlayed, IsDealer, IsRestart);

   /* look for go/lone play/winning play - negative return means continue  */ 
   n = QuickChk(&PegData, OwnPlayed, 121-OwnScr);
   if (n >= 0) return n;

   /* get risk in range [-1.0,1.0] */
   PegData.Risk =
      EvalPegRisk(Cards, OwnPlayed, OppPlayed, OwnScr, OppScr, IsDealer);

   /* load opponent 4-card hand frequencies by rank from GEN.C data - this */
   /* table, that varies according to 7 known cards, replaces static average */
   /* pone and dealer frequency tables used in 1997 version of program */
   LoadHnTot(Cards, HnFreq, IsDealer);

   /* uncomment to generate frequency stats for each hand to file */
   //if (OwnPlayed == 0) DebugDump(Cards, HnFreq, IsDealer);
 
   /* load 13-rank weights array with likelihood opponent still holds rank */
   InitWts(WtAry, HnFreq, Cards+7, OppPlayed, PegData.OppMin);
   
   /* make heuristic adjustments (not from GEN.C) to weights array */
   if (OppPlayed == 3) AdjFlWts(WtAry, Cards);
   AdjustWts(WtAry, Cards, PlAry, OwnPlayed+OppPlayed);
   NormWts(WtAry, PegData.OppMin);

   PickPeg(&PegData, 1);              /* ignore returned weight */;
   n = ConvPeg(Cards, PegData.Bits);  /* convert Bits to index 0-3 */

   return n+1;  /* make 1-based, since go is 0 */
  }

/*----------------------------------------------------------------------------
; Convert or'ed indexes 8/4/2/1 to one card index 0-3, breaking ties by flush
; faking potential.  Assume at least one bit set.
;---------------------------------------------------------------------------*/
int ConvPeg(char Cards[], int Bits)
  {int Mx = 0, MxNdx = 0, n, m, Suit, Sum;

   for (n = 0; n < 4; n++) if (Bits & (8 >> n))
     {Suit = SuitOf(Cards[n]); 
      
      /* adding extra 1 here (i.e., m == 0 below) is bias toward suit of */
      /* first card, since that card may have been played already - played */
      /* cards are shifted to left in hand passed to DLL during pegging */
      for (Sum = m = 0; m < 4; m++) 
         if (Suit == SuitOf(Cards[m])) 
            Sum += 2 + (m == 0); 

      if (Sum > Mx) (Mx = Sum, MxNdx = n);
     }

   return MxNdx;
  }

/*----------------------------------------------------------------------------
; Build count array (counts of cards remaining by rank).
;---------------------------------------------------------------------------*/
void BuildCt(char Cards[], int Length, int CtAry[])
  {int n;

   for (n = 0; n < 16; n++) CtAry[n] = 4;
   for (n = 0; n < Length; n++) --CtAry[RankOf(Cards[n])];
  }

/*----------------------------------------------------------------------------
; Build played card array.  Assume already checked as legal.
;---------------------------------------------------------------------------*/
void BuildPl(LPSTR Cards, int OwnPlayed, int OppPlayed, int IsOwnTurn, LPSTR PlAry)
  {int n, PegCount = 0, GoCnt = 0;
   HANDSTRUCT Own, Opp;
   LPHANDSTRUCT Cur;

   Own.Hand = Cards;   Own.Len = OwnPlayed; Own.Ndx = 0;
   Opp.Hand = Cards+7; Opp.Len = OppPlayed; Opp.Ndx = 0;


   while (Own.Ndx < Own.Len || Opp.Ndx < Opp.Len)
     {Cur = IsOwnTurn ? &Own : &Opp;
      if (Cur->Ndx >= Cur->Len) n = 32;
      else n = PegCount + CountOf(RankOf(Cur->Hand[Cur->Ndx]));

      if (n > 31) ++GoCnt;
      else
        {PegCount = n; GoCnt = 0; *(PlAry++) = Cur->Hand[Cur->Ndx++];
        }

      if (PegCount == 31 || GoCnt > 1) PegCount = GoCnt = 0;

      IsOwnTurn = !IsOwnTurn;
     }
  }

/*----------------------------------------------------------------------------
; Replay pegging to set several elements of peg structure.
;---------------------------------------------------------------------------*/
void Replay(LPPEGSTRUCT PegPtr, int OwnPlayed, int OppPlayed, 
            int IsDealer, int IsRestart)
  {int n, NewCnt, PlayedLen = OwnPlayed + OppPlayed,
       PegCnt = 0, PlCnt = 0, OppMin = 1, GoFlgs = IsDealer ? 1 : 0;
   LPSTR Cards = PegPtr->HnPtr, Played = PegPtr->PlPtr;


   for (n = 0; n < PlayedLen; n++)
     {NewCnt = PegCnt + CountOf(RankOf(Played[n]));

      /* track last two plays--set bit flags own play */
      GoFlgs <<= 1;
      if (Played[n] == *Cards) (Cards++, GoFlgs++);
      GoFlgs &= 3;

      /* update opponent minimum card */
      if ((NewCnt <= 31 && GoFlgs == 3) || NewCnt > 31)
         OppMin = max(OppMin, 32 - PegCnt);


      /* update peg count and card count since last restart */
      if (NewCnt < 31)      (PegCnt = NewCnt, ++PlCnt);
      else if (NewCnt == 31) PegCnt = PlCnt = 0;
      else                  (PegCnt = NewCnt-PegCnt, PlCnt = 1);
     }

   /* handle restart flag if peg count not already zero */
   if (IsRestart && PegCnt)
     {OppMin = max(OppMin, 32 - PegCnt);
      PlCnt = PegCnt = GoFlgs = 0;
     }

   /* update peg structure */
   PegPtr->GoFlgs = GoFlgs & 1;  /* low bit is opponent go */
   PegPtr->PlCnt  = PlCnt;
   PegPtr->PegCnt = PegCnt;
   PegPtr->OppMin = min(OppMin, 10);
   PegPtr->PlPtr += PlayedLen - PlCnt;
  }

/*----------------------------------------------------------------------------
; Look for go, lone play, or winning play.  Return -1 if none of these, else
; 0 for go, or card index 1-4 for play.
;---------------------------------------------------------------------------*/
int QuickChk(LPPEGSTRUCT PegPtr, int OwnPlayed, int PointsToWin)
  {int n, PlayNdx = 0, PlayTot = 0,
       PegCnt = PegPtr->PegCnt, PlCnt = PegPtr->PlCnt;
   LPSTR Cards = PegPtr->HnPtr, Played = PegPtr->PlPtr;

   /* PlayNdx remains 0 if go, but early return if winning play */
   for (n = OwnPlayed; n < 4; n++)
      if (PegCnt + CountOf(RankOf(Cards[n])) <= 31)
        {++PlayTot;
         PlayNdx = n+1;
         Played[PlCnt] = Cards[n];
         if (ScorePeg(Played, PlCnt+1) >= PointsToWin) return PlayNdx;
        }

   return PlayTot > 1 ? -1 : PlayNdx;
  }

/*----------------------------------------------------------------------------
; Bump HnFreq array by scaled weights for opponent held cards.  This is core 
; routine in pegging section, so two different speed optimizations are used.
;
; First, the index pointer array has been precomputed.  Second, the short loop
; has been unrolled as a switch statement.  See similar BumpTot().  
;---------------------------------------------------------------------------*/
void BumpHnTot(int Wt, int MaxWt, int Len, int *Ndx, double HnFreq[], int IsDlr)
  {double Scale;
   short *HnPtr, *NdxPtr;

   /* bump held hand frequency totals by scaled weight */   
   Scale = (double)Wt / MaxWt;

   NdxPtr = HnNdxTbl + *Ndx;
   HnPtr  = (IsDlr ? PnTbl : DlTbl) + *Ndx;

   /* following is unrolled loop */
   switch (Len)
     {
      case 15:
         if (HnPtr[14]) HnFreq[(NdxPtr[14])] += HnPtr[14] * Scale;
         if (HnPtr[13]) HnFreq[(NdxPtr[13])] += HnPtr[13] * Scale;
         if (HnPtr[12]) HnFreq[(NdxPtr[12])] += HnPtr[12] * Scale;
         if (HnPtr[11]) HnFreq[(NdxPtr[11])] += HnPtr[11] * Scale;
      case 11:
         if (HnPtr[10]) HnFreq[(NdxPtr[10])] += HnPtr[10] * Scale;
         if (HnPtr[ 9]) HnFreq[(NdxPtr[ 9])] += HnPtr[ 9] * Scale;
         if (HnPtr[ 8]) HnFreq[(NdxPtr[ 8])] += HnPtr[ 8] * Scale;
      case 8:
         if (HnPtr[ 7]) HnFreq[(NdxPtr[ 7])] += HnPtr[ 7] * Scale;
      case 7:
         if (HnPtr[ 6]) HnFreq[(NdxPtr[ 6])] += HnPtr[ 6] * Scale;
      case 6:
         if (HnPtr[ 5]) HnFreq[(NdxPtr[ 5])] += HnPtr[ 5] * Scale;
      case 5:
         if (HnPtr[ 4]) HnFreq[(NdxPtr[ 4])] += HnPtr[ 4] * Scale;
      case 4:
         if (HnPtr[ 3]) HnFreq[(NdxPtr[ 3])] += HnPtr[ 3] * Scale;
      case 3:
         if (HnPtr[ 2]) HnFreq[(NdxPtr[ 2])] += HnPtr[ 2] * Scale;
         if (HnPtr[ 1]) HnFreq[(NdxPtr[ 1])] += HnPtr[ 1] * Scale;
         if (HnPtr[ 0]) HnFreq[(NdxPtr[ 0])] += HnPtr[ 0] * Scale;
         break;

      default: ;
         /* does not occur--semi-colon needed else omit default */
     }

   *Ndx += Len;
  }

/*----------------------------------------------------------------------------
; Load HnFreq array with weights for held 4-card hands implied by discards 
; from all 6-card hands out of 52-card deck with 7 cards missing.  If dealer 
; flagged, then pone frequencies are generated.  Otherwise dealer frequencies
; are generated.  See similar LoadTot().  The on-the-fly HnFreq table, which 
; varies depending on the 7 missing cards, replaces static average tables used 
; in original version of Cribbage '97.
;---------------------------------------------------------------------------*/
void LoadHnTot(char Hand[], double HnFreq[], int IsDlr)
  {int u, v, w, x, y, z, n, Wt, Ndx, WtRow[16];

   /* initialize frequency array */
   for (n = 0; n < 1820; n++) HnFreq[n] = 0.0;

   /* initialize rank weight counts - Hand[] here includes suits */
   for (n = 1; n < 14; n++) WtRow[n] = 21;
   for (n = 0; n < 7; n++) 
      if (WtRow[RankOf(Hand[n])] > 1) WtRow[RankOf(Hand[n])] -= 5;

   /* update sum arrays for type 111111 hands */
   Ndx = 0;
   for (u = 13;  u > 5; u--)
   for (v = u-1; v > 4; v--)
   for (w = v-1; w > 3; w--)
   for (x = w-1; x > 2; x--)
   for (y = x-1; y > 1; y--)
   for (z = y-1; z > 0; z--)
     {
      Wt = Wgts[WtRow[u]] * Wgts[WtRow[v]] * Wgts[WtRow[w]] *
           Wgts[WtRow[x]] * Wgts[WtRow[y]] * Wgts[WtRow[z]];        /* 111111 */
      BumpHnTot(Wt, 4096, 15, &Ndx, HnFreq, IsDlr);
     }

   /* update sum arrays for type 11112 hands */
   for (u = 13;  u > 4; u--)
   for (v = u-1; v > 3; v--)
   for (w = v-1; w > 2; w--)
   for (x = w-1; x > 1; x--)
   for (y = x-1; y > 0; y--)
     {
      /* 5 positions for the pair */
      Wt = Wgts[1+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[  WtRow[x]] * Wgts[  WtRow[y]];                     /* 21111 */
      BumpHnTot(Wt, 1536, 11, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[  WtRow[x]] * Wgts[WtRow[y]];                       /* 12111 */
      BumpHnTot(Wt, 1536, 11, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[1+WtRow[w]] *
           Wgts[  WtRow[x]] * Wgts[  WtRow[y]];                     /* 11211 */
      BumpHnTot(Wt, 1536, 11, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[1+WtRow[x]] * Wgts[  WtRow[y]];                     /* 11121 */
      BumpHnTot(Wt, 1536, 11, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[  WtRow[x]] * Wgts[1+WtRow[y]];                     /* 11112 */
      BumpHnTot(Wt, 1536, 11, &Ndx, HnFreq, IsDlr);
     }

   /* update sum arrays for type 1122 and 1113 hands */
   for (u = 13;  u > 3; u--)
   for (v = u-1; v > 2; v--)
   for (w = v-1; w > 1; w--)
   for (x = w-1; x > 0; x--)
     {
      /* 1122 hands--6 positions for the 2 pairs */
      Wt = Wgts[1+WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[  WtRow[x]];                                        /* 2211 */
      BumpHnTot(Wt, 576, 8, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[1+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[1+WtRow[w]] *
           Wgts[  WtRow[x]];                                        /* 2121 */
      BumpHnTot(Wt, 576, 8, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[1+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[1+WtRow[x]];                                        /* 2112 */
      BumpHnTot(Wt, 576, 8, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[1+WtRow[w]] *
           Wgts[  WtRow[x]];                                        /* 1221 */
      BumpHnTot(Wt, 576, 8, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[1+WtRow[x]];                                        /* 1212 */
      BumpHnTot(Wt, 576, 8, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[1+WtRow[w]] *
           Wgts[1+WtRow[x]];                                        /* 1122 */
      BumpHnTot(Wt, 576, 8, &Ndx, HnFreq, IsDlr);

      /* 1113 hands--4 positions for the trips */
      Wt = Wgts[2+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[  WtRow[x]];                                        /* 3111 */
      BumpHnTot(Wt, 256, 7, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[2+WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[  WtRow[x]];                                        /* 1311 */
      BumpHnTot(Wt, 256, 7, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[2+WtRow[w]] *
           Wgts[  WtRow[x]];                                        /* 1131 */
      BumpHnTot(Wt, 256, 7, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]] *
           Wgts[2+WtRow[x]];                                        /* 1113 */
      BumpHnTot(Wt, 256, 7, &Ndx, HnFreq, IsDlr);
     }

   /* update sum arrays for type 123, 114, and 222 hands */
   for (u = 13;  u > 2; u--)
   for (v = u-1; v > 1; v--)
   for (w = v-1; w > 0; w--)
     {
      /* 123 hands--6 positions for pair/trips */
      Wt = Wgts[2+WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[  WtRow[w]];  /* 321 */
      BumpHnTot(Wt, 96, 5, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[2+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[1+WtRow[w]];  /* 312 */
      BumpHnTot(Wt, 96, 5, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[1+WtRow[u]] * Wgts[2+WtRow[v]] * Wgts[  WtRow[w]];  /* 231 */
      BumpHnTot(Wt, 96, 5, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[1+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[2+WtRow[w]];  /* 213 */
      BumpHnTot(Wt, 96, 5, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[2+WtRow[v]] * Wgts[1+WtRow[w]];  /* 132 */
      BumpHnTot(Wt, 96, 5, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[2+WtRow[w]];  /* 123 */
      BumpHnTot(Wt, 96, 5, &Ndx, HnFreq, IsDlr);

      /* 114 hands--3 positions for quad */
      Wt = Wgts[3+WtRow[u]] * Wgts[  WtRow[v]] * Wgts[  WtRow[w]];  /* 411 */
      BumpHnTot(Wt, 16, 4, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[3+WtRow[v]] * Wgts[  WtRow[w]];  /* 141 */
      BumpHnTot(Wt, 16, 4, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[  WtRow[u]] * Wgts[  WtRow[v]] * Wgts[3+WtRow[w]];  /* 114 */
      BumpHnTot(Wt, 16, 4, &Ndx, HnFreq, IsDlr);

      /* 222 hands--1 position for 3 pairs */
      Wt = Wgts[1+WtRow[u]] * Wgts[1+WtRow[v]] * Wgts[1+WtRow[w]];  /* 222 */
      BumpHnTot(Wt, 216, 6, &Ndx, HnFreq, IsDlr);
     }

   /* update sum arrays for type 24 and 33 hands */
   for (u = 13;  u > 1; u--)
   for (v = u-1; v > 0; v--)
     {
      /* 24 hands--2 positions for pair/quad */
      Wt = Wgts[3+WtRow[u]] * Wgts[1+WtRow[v]];                     /* 42 */
      BumpHnTot(Wt, 6, 3, &Ndx, HnFreq, IsDlr);

      Wt = Wgts[1+WtRow[u]] * Wgts[3+WtRow[v]];                     /* 24 */
      BumpHnTot(Wt, 6, 3, &Ndx, HnFreq, IsDlr);

      /* 33 hands--1 position for 2 trips */
      Wt = Wgts[2+WtRow[u]] * Wgts[2+WtRow[v]];                     /* 33 */
      BumpHnTot(Wt, 16, 3, &Ndx, HnFreq, IsDlr);
     }

   /* normalize sum arrays to frequencies */
   for (n = 0; n < 1820; n++) HnFreq[n] /= SUM45;
  }

/*----------------------------------------------------------------------------
; From GEN.C 1820-entry opponent frequency table, known opponent ranks, and  
; known opponent rank minimum, populate 13-entry weights table with estimated 
; frequency distribution of remaining unknown ranks.
;---------------------------------------------------------------------------*/
void InitWts(double WtAry[], double FreqTbl[], char KnownCds[], int Len, int RankMin)
  {int n, m, r1, r2, r3, r4, Found, Rank, Ranks[4], List[4];
   double Wt;

   /* initialize weights with a positive value less than any one addition below */    
   for (n = RankMin; n < 14; n++) WtAry[n] = 0.0000001;

   /* fix array of known opponent ranks */
   for (n = 0; n < Len; n++) Ranks[n] = RankOf(KnownCds[n]);
   
   /* early out if nothing else would be added below */
   if (Len == 4) return;
   
   /* about 20% of time with 3 known ranks, no weight gets added to weights */
   /* array, so rely in this case on subsequent heuristic scaling of the small */
   /* positive initial values--scaling would not work with zero initialization */
   
   /* loop through 1820 frequency entries */
   for (r1 = 1; r1 <= 13; r1++)
   for (r2 = 1; r2 <= r1; r2++)
   for (r3 = 1; r3 <= r2; r3++)
   for (r4 = 1; r4 <= r3; r4++)
   if ((Wt = *(FreqTbl++)))  /* assignment (and insure bump before looping) */
   if (r4 >= RankMin)        /* r4 <= r3 <= r2 <= r1 */
     {
      /* load ranks to temp array set */
      List[0] = r1; List[1] = r2; List[2] = r3; List[3] = r4;

      /* insure all known ranks have match in current rank set */
      Found = -1; n = 0;
      while (Found && n < Len)
        {Rank = Ranks[n++];
         Found = 0;
         for (m = 0; m < 4; m++) if (Rank == List[m])
           {List[m] = 0;  /* prevent double hits and WtAry bump below */
            Found = -1; 
            break;
           }
        }

      /* if so (or if no known ranks), contribute for missing ranks */
      if (Found)
        {if (List[0]) WtAry[r1] += Wt;
         if (List[1]) WtAry[r2] += Wt;
         if (List[2]) WtAry[r3] += Wt;
         if (List[3]) WtAry[r4] += Wt;
        }
     }

   /* normalize later */
  }

/*----------------------------------------------------------------------------
; Adjust weight array for flush potential if 3 opponent plays all same suit.
;---------------------------------------------------------------------------*/
#define DROP_FAC1 0.5

void AdjFlWts(double WtAry[], char Cards[])
  {int n, Suit;

   /* ignore if not same suit or rank without suit */
   Suit = SuitOf(Cards[7]) & Cards[8] & Cards[9];
   if (Suit == 0) return;

   /* scan all known cards (own/cut/opp) for suit match */
   for (n = 0; n < 10; n++) if (Suit == SuitOf(Cards[n]))
      WtAry[RankOf(Cards[n])] *= DROP_FAC1;
  }

/*----------------------------------------------------------------------------
; Reduce weights of possible scoring cards when actual play did not score.
;---------------------------------------------------------------------------*/
#define DROP_FAC2 0.6

void DropWts(double WtAry[], LPSTR PlAry, int Length, int PegCnt)
  {int n;
   char Save;
   LPSTR Last = PlAry+Length-1;

   /* if scoring play, do nothing */
   if (ScorePeg(PlAry, Length)) return;

   /* make temp substitutions to find declined scoring plays */
   Save = *Last;
   for (n = 1; n < 14; n++)
     if (n != RankOf(Save) && PegCnt + CountOf(n) <= 31)
        if (*Last = (char)n, ScorePeg(PlAry, Length))
           WtAry[n] *= DROP_FAC2;
   *Last = Save;
  }

/*----------------------------------------------------------------------------
; Fear factor decreases for each played card as play proceeds. 1.0 is neutral. 
;---------------------------------------------------------------------------*/
#define FEAR_FAC1  2.4
#define FEAR_FAC2  1.6

const double ScaleFac[] = {FEAR_FAC1, FEAR_FAC2, 1.0, 1.0};

/*----------------------------------------------------------------------------
; Adjust weight array down for declined scoring plays and up for cards played.
;---------------------------------------------------------------------------*/
void AdjustWts(double WtAry[], char OwnCards[], char Played[], int Length)
  {int n, m, OppCnt = 0, PegCnt = 0, PlCnt = 0, NewCnt, Found = 0;
   char Card;

   for (n = 0; n < Length; n++)
     {Card = Played[n];
      NewCnt = PegCnt + CountOf(RankOf(Card));

      /* check if played card is one of own cards */
      for (m = 0; m < 4; m++) if (Card == OwnCards[m]) break;
      
      /* if opponent card (no early break) */
      if (m == 4)
        {
         /* fear repeat card by opponent */
         WtAry[RankOf(Card)] *= ScaleFac[OppCnt++ & 3];

         /* account for declined scoring plays */
         if (NewCnt <= 31) DropWts(WtAry, Played+n-PlCnt, PlCnt+1, PegCnt);
        }

      if (NewCnt > 31)       (PegCnt = NewCnt - PegCnt, PlCnt = 1);
      else if (NewCnt == 31)  PegCnt = PlCnt = 0;
      else                   (PegCnt = NewCnt, ++PlCnt);
     }

  }

/*----------------------------------------------------------------------------
; Normalize weights array (13 or fewer ranks).
;---------------------------------------------------------------------------*/
void NormWts(double WtAry[], int RankMin)
  {int n;
   double Wt = 0.0;

   /* normalize weight results */
   for (n = RankMin; n < 14; n++) Wt += WtAry[n];
   if (Wt > 0.0) for (n = RankMin; n < 14; n++) WtAry[n] /= Wt;
  }

/*----------------------------------------------------------------------------
; Return (x/y)**z, where z is non-negative integer and x/y in range [0.0,1.0].
;---------------------------------------------------------------------------*/
double GetProb(double x, double y, int z)
  {
   if (x == 0.0 || y == 0.0) return 0.0;
   x = y = x / y;
   if (x >= 1.0 || z == 0)   return 1.0;
   while (--z) x *= y;

   return x;
  }

/*----------------------------------------------------------------------------
; Return (x/y)*z, where x/y expected (forced) in range [0.0,256.0] and z in 
; range [0.0,1.0].
;---------------------------------------------------------------------------*/
double ProdDiv(double x, double y, double z)
  {
   if (x == 0.0 || y == 0.0) return 0.0;
   x = x / y; 
   if (x > 256.0) x = 256.0;
   if (x <   0.0) x =   0.0;

   return  x * z;
  }

/*----------------------------------------------------------------------------
; Excess-128 sum values in range [0.0,256.0] with overflow/underflow caps.  
; This was done on 2019 upgrade from byte to double.  Byte held estimated 
; scoring in 1/8 point increments, with 0x80 being neutral.  Low value
; favored opponent and high value favored player.
;---------------------------------------------------------------------------*/
double XsAdd(double x, double y)
  {
   x += y - 128.0;
   if (x > 256.0) x = 256.0;
   if (x <   0.0) x =   0.0;

   return x;
  }

/*----------------------------------------------------------------------------
; From pointer to peg structure, set best peg choices in Bits field and return
; evaluation weight from function.  Peg choices are or'ed indexes 8/4/2/1, or  
; 0 for go.  Data unchanged on exit except free tail of played card array, 
; free tail of cards array, and own GoFlgs bit.
;---------------------------------------------------------------------------*/
#define OPP_GO  1
#define OWN_GO  2
#define BOTH_GO 3

double PickPeg(LPPEGSTRUCT PegPtr, int IsTopLevel)
  {int n, Bits = 0, Mask;
   double Wt, MxWt = 0.0;
   char Card;
   double Ave[] = {0.0, 0.0, 0.0, 0.0};  /* debugging info */

   /* insure own go flag clear */
   PegPtr->GoFlgs &= OWN_GO ^ 0xFF;

   /* weigh available card plays within 31-limit */
   for (n = 0; n < 4; n++)
     {Card = PegPtr->HnPtr[n];
      Mask = 8 >> n;
      if (PegPtr->HnFlgs & Mask)                           /* unplayed */
         if (PegPtr->PegCnt + CountOf(RankOf(Card)) <= 31) /* playable */
            {PegPtr->PlPtr[PegPtr->PlCnt] = Card;
             PegPtr->HnFlgs ^= Mask;                 /* flag as played */
             Wt = PegAve(*PegPtr);                   /* RECURSION */
             Ave[n] = Wt;                            /* debugging info */
             PegPtr->HnFlgs |= Mask;                 /* flag as unplayed */
             if (Wt > MxWt) (MxWt = Wt, Bits = Mask);
             else if (Wt == MxWt) Bits |= Mask;
            }
     }

   /* if no plays, then evaluate go instead */
   if (Bits == 0)
     {PegPtr->GoFlgs |= OWN_GO;
      MxWt = PegAve(*PegPtr);                        /* RECURSION */
     }

   /* after top level call, Bits hold final choice(s) */
   PegPtr->Bits = Bits;

   /* function called at top level PickPeg() when line uncommented */
   //if (IsTopLevel && DebugWanted(PegPtr)) DebugPlay(PegPtr, Ave);

   return MxWt;
  }

/*----------------------------------------------------------------------------
; Return peg weight of card at end of played array or of go if flagged in
; GoFlgs.  Assume HnFlgs already updated.  High return value is good for 
; player.  Low return value is good for opponent.  Player/opponent peg weights
; are in 1/8 points, but with opponent scores scaled up or down based on risk. 
; and likelihood opponent has the scoring card.  Result in range [0.0,256.0] 
; where 128.0 is neutral (player scoring balanced by scaled opponent scoring).
;---------------------------------------------------------------------------*/
double PegAve(PEGSTRUCT PegData)    /* copy of structure, not pointer */
  {double Wt = 128.0, Scale = 1.0;  /* neutral weight and scale initially */

   /* most of time, scale will be neutral 1.0, but when risk near -1.0 or 1.0,
   /* opponent scale factor set to lower or higher value */
   Scale = GetPegScale(PegData.Risk);

   /* initial scoring here is directly added to or subtracted from Wt, but  */
   /* this shortcut is the same as Wt = XsAdd(Wt, 128.0 + <player eighths>) */
   /* or Wt = XsAdd(Wt, 128.0 - <scaled opponent eighths>), similar to  */
   /* code further below when deducting opponent eighths from PickWt */
 
   /* evaluate card at end of played cards with weight in 1/8th points */
   if ((PegData.GoFlgs & OWN_GO) == 0)
     {PegData.PegCnt += CountOf(RankOf(PegData.PlPtr[PegData.PlCnt]));
      PegData.PlCnt++;
      Wt += 8.0 * ScorePeg(PegData.PlPtr, PegData.PlCnt);    /* own score */
     }
   else if (PegData.GoFlgs & OPP_GO) Wt += 8.0;           /* own go point */

   /* handle reset for go's/31, as needed */
   if ((PegData.GoFlgs & BOTH_GO) == BOTH_GO || PegData.PegCnt == 31)
      {PegData.PlPtr += PegData.PlCnt;
       PegData.PegCnt = PegData.GoFlgs = PegData.PlCnt = 0;
      }

   /* handle termination when cards gone, including last go check */
   if (PegData.HnFlgs == 0 && PegData.OppLft == 0)
     {if (PegData.PegCnt)
        {
         if ((PegData.GoFlgs & OWN_GO) == 0) Wt += 8.0;  /* own go point */
         else Wt -= 8.0 * Scale;                         /* opp go point */
        }
      return Wt;                                    /* RECURSION TERMINATION */
     }

   /* handle sure opponent go--called or implied */
   if ((PegData.GoFlgs & OPP_GO) ||
        PegData.OppLft == 0      ||
        PegData.PegCnt + PegData.OppMin > 31)
     {
      PegData.GoFlgs |= OPP_GO;
      Wt = XsAdd(Wt, PickPeg(&PegData, 0));         /* RECURSION */
      return Wt;                                    /* OPPONENT GO EXIT */
     }

   /* check depth */
   if (PegData.Depth == 0) return Wt;               /* RECURSION TERMINATION */

   /* at this point, own play or go evaluated, recursion termination checked */
   /* and opponent has go or non-go plays, which must be evaluated. */

  {int n, SvPlCnt = PegData.PlCnt, SvPegCnt = PegData.PegCnt, Pts;
   double GoWt = 0.0, PtWt = 0.0, ZrWt = 0.0, GoAve = 0.0, CurWt, PickWt,
          PtSum = 0.0, ZrSum = 0.0, OppWt, GoProb, NoProb;
   LPSTR SvPlPtr = PegData.PlPtr;

   /* save Wt and GoAve til end */
   if (SvPegCnt > 21)
     {n = PegData.OppMin;
      PegData.OppMin = 32-SvPegCnt;
      PegData.GoFlgs |= OPP_GO;
      GoAve = PickPeg(&PegData, 0);                 /* RECURSION */
      PegData.OppMin = n;
     }

   /* prepare for loop--other elements changed and reset in loop */
   PegData.Depth--;
   PegData.OppLft--;
   PegData.GoFlgs &= OPP_GO ^ 0xFF;  /* insure opponent go clear */

   /* loop over opponent plays (including those for go evaluation) */
   for (n = PegData.OppMin; n <= 13; n++)
     {
      /* get rank weight--tentatively keep 1997 count reduction although */
      /* new table already reflects some of the reduction */
      CurWt = PegData.WtPtr[n] * PegData.CtPtr[n] / 4.0;

      /* if no weight, loop--if go, bump go weight and loop */
      if (CurWt == 0.0 || PegData.CtPtr[n] == 0) continue; 
      if (SvPegCnt + CountOf(n) > 31) {GoWt += CurWt; continue;}

      /* handle card play */
      SvPlPtr[SvPlCnt] = n;
      Pts = ScorePeg(SvPlPtr, SvPlCnt+1);

      /* set peg structure for next recursion call */
      if (SvPegCnt + CountOf(n) == 31)
        {PegData.PlPtr  = SvPlPtr+SvPlCnt+1;
         PegData.PegCnt = PegData.PlCnt = 0;
        }
      else
        {PegData.PegCnt = SvPegCnt+CountOf(n);
         PegData.PlCnt  = SvPlCnt+1;
        }

      /* recurse for own play, with depth undo if opponent scored */
      if (Pts) PegData.Depth++;
      PegData.CtPtr[n]--;
      PickWt = PickPeg(&PegData, 0);               /* RECURSION */
      PegData.CtPtr[n]++;
      if (Pts) PegData.Depth--;

      /* update sums for scoring and non-scoring plays */
      if (Pts)
        {
         /* reduce own PickWt by the opponent scaled eighth points--with */
         /* scale factor, 14-point pegging play (4-of-a-kind at 31) can */
         /* easily trigger underflow cap in XsAdd() */
         PickWt = XsAdd(PickWt, 128.0 - 8.0 * Pts * Scale);

         PtWt  += CurWt;
         PtSum += CurWt * PickWt;
        }
      else
        {ZrWt  += CurWt;
         ZrSum += CurWt * PickWt;
        }
     } /* end for loop */

   /* get probability bytes--scoring/zero score probabilities implied */
   n = PegData.OppLft+1;
   GoProb = GetProb(GoWt,      GoWt+ZrWt+PtWt, n);  /* go prob */
   NoProb = GetProb(GoWt+ZrWt, GoWt+ZrWt+PtWt, n);  /* no score prob */

   /* sum three contributions to opponent score */
   OppWt  = GoProb * GoAve;                       /* go contribution */
   OppWt += ProdDiv(ZrSum, ZrWt, NoProb-GoProb);  /* zero score contribution */
   OppWt += ProdDiv(PtSum, PtWt, 1.0-NoProb);     /* scoring contribution */

   /* combine own and opponent contributions */
   Wt = XsAdd(Wt, OppWt);

  } /* end extra declarations section */

   return Wt;
  }

/****************************************************************************/
/************************** Debugging functions *****************************/
/****************************************************************************/

/*----------------------------------------------------------------------------
; Dump HnFreq array to file.  Show 7 excluded cards, dealer flag, and 4-rank 
; hand associated with each entry.  Entry frequency is a percentage (100 times
; probability).  Maximum entry typically near 1 percent, e.g., QJT5 or 9876. 
;---------------------------------------------------------------------------*/
void DebugDump(char Hand[], double HnFreq[], int IsDlr)
  {int n, r1, r2, r3, r4, Ct[16], Cnt = 0, i = 0;
   double Wt, Mx = 0.0;
   char Hn[7], Rn[4];
   const char Rnk[] = " A23456789TJQK";
   FILE *fp;
   
   // open for append, but create if does not exist
   if ((fp = fopen("DumpTbl.txt", "a")) == NULL) return;

   // get known ranks, zero weight count, and maximum weight
   for (n = 0; n < 7; n++) Hn[n] = Rnk[RankOf(Hand[n])];
   for (n = 0; n < 1820; n++)
      (Wt = HnFreq[n], Cnt += (Wt == 0.0), Mx = (Wt > Mx) ? Wt : Mx);

   // print table header info with zero count/marker descriptions
   fprintf(fp,"Excluded %s cards/discards/cut: %c%c%c%c-%c%c-%c\n",
      IsDlr ? "dealer":"pone", Hn[0],Hn[1],Hn[2],Hn[3],Hn[4],Hn[5],Hn[6]); 
   fprintf(fp,"%s zero percentages (blanks or plus signs): %d of 1820\n",
      IsDlr ? "Pone":"Dealer", Cnt);
   fprintf(fp,"Blank means impossible--plus means negligible\n");
   fprintf(fp,"Maximum %s percentage (asterisks): %6.4f\n\n",
      IsDlr ? "pone":"dealer", 100 * Mx);

   // print 1820 weight percentages (or blank/plus/asterisk markers) 
   for (r1 = 1; r1 <= 13; r1++)
   for (r2 = 1; r2 <= r1; r2++)
   for (r3 = 1; r3 <= r2; r3++)
   for (r4 = 1; r4 <= r3; r4++)
     {if ((Wt = HnFreq[i++]) == 0.0)
        {Rn[0] = r1; Rn[1] = r2; Rn[2] = r3; Rn[3] = r4;

         // distinguish negligible weight from impossible holding (n < 4)
         for (n = 1; n < 14; n++) Ct[n] = 4;
         for (n = 0; n < 7; n++) --Ct[RankOf(Hand[n])];
         for (n = 0; n < 4; n++) if (--Ct[Rn[n]] < 0) break;
        }

      // print holding followed by weight percentage or marker
      fprintf(fp, "%c%c%c%c=", Rnk[r1],Rnk[r2],Rnk[r3],Rnk[r4]);
      if (Wt == 0.0 && n == 4) fprintf(fp, "++++++");
      else if (Wt == 0.0)      fprintf(fp, "      ");
      else if (Wt < Mx)        fprintf(fp, "%6.4f", 100 * Wt);
      else                     fprintf(fp, "******");
      fprintf(fp, "%c", (i % 8) ? ' ' : '\n'); 
     }
   fprintf(fp, "\n\n");

   fclose(fp);
  }

/*----------------------------------------------------------------------------
; Return true when want pegging debug display at top level.  Edit as needed. 
;---------------------------------------------------------------------------*/
int DebugWanted(LPPEGSTRUCT PegPtr)
  {int PnRank, DlRank, n;

   // limit to first peg play as dealer
   if (PegPtr->HnFlgs != 15 || PegPtr->OppLft != 3) return 0;

   // limit to dealer hands where pairing pone is option
   PnRank = RankOf(PegPtr->PlPtr[0]);
   for (n = 0; n < 4; n++) if (RankOf(PegPtr->HnPtr[n]) == PnRank) break;
   if (n == 4) return 0;

   // limit to declined scoring play (no pair, no 15)
   n = ConvPeg(PegPtr->HnPtr, PegPtr->Bits);
   DlRank = RankOf(PegPtr->HnPtr[n]);
   if (PnRank == DlRank || CountOf(PnRank) + CountOf(DlRank) == 15) return 0; 

   return 1;
  }

/*----------------------------------------------------------------------------
; Display debugging info in top level PickPeg(), if call there uncommented. 
;---------------------------------------------------------------------------*/
void DebugPlay(LPPEGSTRUCT PegPtr, double Ave[])
  {int n, a[13];
   char Buf[300], Cds[40], Pls[30], Wts[70], Cts[40], Avs[40], Misc[60];
   const char Rnk[] = " A23456789TJQK";

   for (n=0; n<13; n++) a[n]=32;
   for (n=0; n<11-PegPtr->OppLft; n++) a[n]=Rnk[RankOf(PegPtr->HnPtr[n])];
   sprintf(Cds, "Cards:\t%c %c %c %c - %c %c - %c - %c %c %c %c\n",
      a[0],a[1],a[2],a[3],a[4],a[5],a[6],a[7],a[8],a[9],a[10]);

   for (n=0; n<13; n++) a[n]=32;
   for (n=0; n<PegPtr->PlCnt; n++) a[n]=Rnk[RankOf(PegPtr->PlPtr[n])];
   sprintf(Pls, "Played:\t%c %c %c %c %c %c %c %c\n",
      a[0],a[1],a[2],a[3],a[4],a[5],a[6],a[7]);

   // scale up weights as much as can while likely avoiding line wrap
   for (n=0; n<13; n++) a[n]=(int)(1300*PegPtr->WtPtr[n+1]);
   sprintf(Wts, "Weights:\t%u %u %u %u %u - %u %u %u %u %u - %u %u %u\n",
      a[0],a[1],a[2],a[3],a[4],a[5],a[6],a[7],a[8],a[9],a[10],a[11],a[12]);

   for (n=0; n<13; n++) a[n]=PegPtr->CtPtr[n+1];
   sprintf(Cts, "Counts:\t%u %u %u %u %u - %u %u %u %u %u - %u %u %u\n",
      a[0],a[1],a[2],a[3],a[4],a[5],a[6],a[7],a[8],a[9],a[10],a[11],a[12]);

   sprintf(Avs, "Aves:\t%.2f  %.2f  %.2f  %.2f\n",Ave[0],Ave[1],Ave[2],Ave[3]);

   // add one to risk to avoid minus sign--display range is now [0.0,2.0]
   a[0]=PegPtr->Depth; a[1]=PegPtr->PegCnt; a[2]=PegPtr->GoFlgs;
   a[3]=PegPtr->Bits; a[4]=PegPtr->OppMin; 
   sprintf(Misc, "Risk: %.4f  Depth/PgCnt/GoFlgs/Bits/OppMn: %u/%u/%u/%u/%u\n\n",
      1+PegPtr->Risk,a[0],a[1],a[2],a[3],a[4]);

   sprintf(Buf, "%s%s%s%s%s%s", Misc, Cds, Pls, Avs, Wts, Cts);
   MessageBox(GetActiveWindow(), Buf, "DebugPlay", MB_ICONINFORMATION);
  }

/****************************************************************************/
/*********************************  EOF  ************************************/
/****************************************************************************/
